import awkward1
from coffea.nanoevents.methods.util import apply_global_index
from coffea.nanoevents.methods.mixin import mixin_class, mixin_method
from coffea.nanoevents.methods.base import NanoCollection
from coffea.nanoevents.methods.vector import PtEtaPhiMLorentzVector
from coffea.nanoevents.methods.candidate import PtEtaPhiMCandidate


@mixin_class
class GenParticle(PtEtaPhiMLorentzVector, NanoCollection):
    """NanoAOD generator-level particle object, including parent and child self-references

    Parent and child self-references are constructed from the ``genPartIdxMother`` column, where
    for each entry, the mother entry index is recorded, or -1 if no mother exists.
    """

    FLAGS = [
        "isPrompt",
        "isDecayedLeptonHadron",
        "isTauDecayProduct",
        "isPromptTauDecayProduct",
        "isDirectTauDecayProduct",
        "isDirectPromptTauDecayProduct",
        "isDirectHadronDecayProduct",
        "isHardProcess",
        "fromHardProcess",
        "isHardProcessTauDecayProduct",
        "isDirectHardProcessTauDecayProduct",
        "fromHardProcessBeforeFSR",
        "isFirstCopy",
        "isLastCopy",
        "isLastCopyBeforeFSR",
    ]
    """bit-packed statusFlags interpretations.  Use `GenParticle.hasFlags` to query"""

    def hasFlags(self, *flags):
        """Check if one or more status flags are set

        Parameters
        ----------
            flags : str or list
                A list of flags that are required to be set true. If the first argument
                is a list, it is expanded and subsequent arguments ignored.
                Possible flags are enumerated in the `FLAGS` attribute

        Returns a boolean array
        """
        if not len(flags):
            raise ValueError("No flags specified")
        elif isinstance(flags[0], list):
            flags = flags[0]
        mask = 0
        for flag in flags:
            mask |= 1 << self.FLAGS.index(flag)
        return (self.statusFlags & mask) == mask

    @property
    def parent(self):
        """Accessor to the parent particle"""
        return apply_global_index(self.genPartIdxMotherG, self._events().GenPart)

    @property
    def distinctParent(self):
        return apply_global_index(self.distinctParentIdxG, self._events().GenPart)

    @property
    def children(self):
        return apply_global_index(self.childrenIdxG, self._events().GenPart)

    @property
    def distinctChildren(self):
        return apply_global_index(self.distinctChildrenIdxG, self._events().GenPart)


@mixin_class
class GenVisTau(PtEtaPhiMCandidate, NanoCollection):
    """NanoAOD visible tau object"""

    @property
    def parent(self):
        """Accessor to the parent particle"""
        return apply_global_index(self.genPartIdxMotherG, self._events().GenPart)



from functools import wraps
import numpy
import awkward1
from coffea.nanoevents.methods.mixin import mixin_class, mixin_method
from coffea.nanoevents.factory import NanoEventsFactory


@mixin_class
class NanoEvents:
    """NanoEvents mixin class"""

    @property
    def metadata(self):
        return self.layout.parameter("metadata")


@mixin_class
class NanoCollection:
    def _getlistarray(self):
        """Do some digging to find the initial listarray"""

        def descend(layout, depth):
            islistarray = isinstance(
                layout,
                (awkward1.layout.ListOffsetArray32, awkward1.layout.ListOffsetArray64,),
            )
            if islistarray and layout.content.parameter("collection_name") is not None:
                return lambda: layout

        return awkward1._util.recursively_apply(self.layout, descend)

    def _starts(self):
        """Internal method to get jagged collection starts

        This should only be called on the original unsliced collection array.
        Used to convert local indexes to global indexes"""
        return numpy.asarray(self._getlistarray().starts).astype("i8")

    def _content(self):
        """Internal method to get jagged collection content

        This should only be called on the original unsliced collection array.
        Used with global indexes to resolve cross-references"""
        return self._getlistarray().content

    def _events(self):
        """Internal method to get the originally-constructed NanoEvents

        This can be called at any time from any collection, as long as
        the NanoEventsFactory instance exists."""
        key = self.layout.purelist_parameter("events_key")
        return NanoEventsFactory.get_events(key)


@mixin_class
class NanoColumn:
    """A column from the original file"""

    pass


def runtime_cache(method):
    """Cache a NanoCollection method in the runtime cache

    This decorator can be placed on any method that is a member
    of a class that subclasses NanoCollection. Such methods should
    only be called on the original unsliced collection array."""

    @wraps(method)
    def cachedmethod(self, *args, **kwargs):
        events_key = self.layout.purelist_parameter("events_key")
        collection_name = self.layout.purelist_parameter("collection_name")
        cache = NanoEventsFactory.get_cache(events_key)
        key = "/".join([events_key, "runtime", collection_name, method.__name__])
        try:
            return cache[key]
        except KeyError:
            return method(*args, **kwargs)

    return cachedmethod



import awkward1
from coffea.nanoevents.methods.util import apply_global_index
from coffea.nanoevents.methods.mixin import mixin_class, mixin_method
from coffea.nanoevents.methods.base import NanoCollection
from coffea.nanoevents.methods.vector import PtEtaPhiMLorentzVector, PolarTwoVector


@mixin_class
class Jet(PtEtaPhiMLorentzVector, NanoCollection):
    """NanoAOD narrow radius jet object"""

    _enable_genjet = False
    "Set to true if using NanoAODv6 or newer (v5 had a bug in the mapping)"
    LOOSE = 0
    "jetId bit position"
    TIGHT = 1
    "jetId bit position"
    TIGHTLEPVETO = 2
    "jetId bit position"

    @property
    def isLoose(self):
        """Returns a boolean array marking loose jets according to jetId index"""
        return (self.jetId & (1 << self.LOOSE)) != 0

    @property
    def isTight(self):
        """Returns a boolean array marking tight jets according to jetId index"""
        return (self.jetId & (1 << self.TIGHT)) != 0

    @property
    def isTightLeptonVeto(self):
        """Returns a boolean array marking tight jets with explicit lepton veto according to jetId index"""
        return (self.jetId & (1 << self.TIGHTLEPVETO)) != 0

    @property
    def matched_electrons(self):
        return apply_global_index(self.electronIdxG, self._events().Electron)

    @property
    def matched_muons(self):
        return apply_global_index(self.muonIdxG, self._events().Muon)

    @property
    def matched_gen(self):
        return apply_global_index(self.genJetIdxG, self._events().GenJet)


@mixin_class
class FatJet(PtEtaPhiMLorentzVector, NanoCollection):
    """NanoAOD large radius jet object"""

    subjetmap = {
        "FatJet": "SubJet"
    }  # V6 has 'GenJetAK8': 'SubGenJetAK8', maybe better to put in generator.py
    LOOSE = 0
    "jetId bit position"
    TIGHT = 1
    "jetId bit position"
    TIGHTLEPVETO = 2
    "jetId bit position"

    @property
    def isLoose(self):
        """Returns a boolean array marking loose jets according to jetId index"""
        return (self.jetId & (1 << self.LOOSE)) != 0

    @property
    def isTight(self):
        """Returns a boolean array marking tight jets according to jetId index"""
        return (self.jetId & (1 << self.TIGHT)) != 0

    @property
    def isTightLeptonVeto(self):
        """Returns a boolean array marking tight jets with explicit lepton veto according to jetId index"""
        return (self.jetId & (1 << self.TIGHTLEPVETO)) != 0

    @property
    def subjets(self):
        return apply_global_index(self.subJetIdxG, self._events().SubJet)


@mixin_class
class MissingET(PolarTwoVector, NanoCollection):
    @property
    def r(self):
        return self["pt"]



"""Awkward 0.x NanoEvents and helpers

See `coffea.nanoevents` for an awkward1-based implementation

This package provides NanoEvents, a utility to wrap the CMS NanoAOD (or similar)
flat nTuple structure into a single awkward array with appropriate object methods
(such as Lorentz vector methods), cross references, and pre-built nested objects,
all lazily accessed from the source ROOT TTree via uproot.

NanoEvents is in a **experimental** stage at this point. Certain functionality may be
fragile, and some functionality will not be available until it is ported to awkward-array version 1.
"""
from .nanoevents import NanoCollection, NanoEvents

__all__ = [
    'NanoCollection',
    'NanoEvents',
]



"""Utility functions

"""
import awkward
import hashlib

from awkward.util import numpy
import numba

akd = awkward
np = numpy
nb = numba

import lz4.frame
import cloudpickle


def load(filename):
    '''Load a coffea file from disk
    '''
    with lz4.frame.open(filename) as fin:
        output = cloudpickle.load(fin)
    return output


def save(output, filename):
    '''Save a coffea object or collection thereof to disk

    This function can accept any picklable object.  Suggested suffix: ``.coffea``
    '''
    with lz4.frame.open(filename, 'wb') as fout:
        thepickle = cloudpickle.dumps(output)
        fout.write(thepickle)


def _hex(string):
    try:
        return string.hex()
    except AttributeError:
        return "".join("{:02x}".format(ord(c)) for c in string)


def _ascii(maybebytes):
    try:
        return maybebytes.decode('ascii')
    except AttributeError:
        return maybebytes


def _hash(items):
    # python 3.3 salts hash(), we want it to persist across processes
    x = hashlib.md5(bytes(';'.join(str(x) for x in items), 'ascii'))
    return int(x.hexdigest()[:16], base=16)



import sys
import os
import os.path as osp


from coffea import hist, processor

try:
    import work_queue as wq
    work_queue_port = 9123
except ImportError:
    print("work_queue is not installed. Omiting test.")
    sys.exit(0)

def template_analysis(environment_file, filelist, executor, flatten, compression):
    from coffea.processor import run_uproot_job
    treename = 'Events'
    from coffea.processor.test_items import NanoTestProcessor

    exe_args = {
        'flatten': flatten,
        'compression': compression,
        'environment-file': environment_file,
        'resources-mode': 'fixed',
        'cores': 2,
        'memory': 500,  # MB
        'disk': 1000,   # MB
        'master-name': 'coffea-test',
        'port': work_queue_port,
        'print-stdout' : True
    }

    hists = run_uproot_job(filelist, treename, NanoTestProcessor(), executor, executor_args = exe_args)

    print(hists)
    assert(hists['cutflow']['ZJets_pt'] == 18)
    assert(hists['cutflow']['ZJets_mass'] == 6)
    assert(hists['cutflow']['Data_pt'] == 84)
    assert(hists['cutflow']['Data_mass'] == 66)

def work_queue_example(environment_file):
    from coffea.processor import work_queue_executor

    # Work Queue does not allow absolute paths
    filelist = {
        'ZJets': ['./samples/nano_dy.root'],
        'Data': ['./samples/nano_dimuon.root']
    }

    workers = wq.Factory(batch_type='local', master_host_port='localhost:{}'.format(work_queue_port))
    workers.max_workers = 1
    workers.min_workers = 1
    workers.cores  = 4
    workers.memory = 1000  # MB
    workers.disk   = 4000  # MB

    with workers:
        #template_analysis(environment_file, filelist, work_queue_executor, flatten=False, compression=0)
        #template_analysis(environment_file, filelist, work_queue_executor, flatten=True, compression=0)
        #template_analysis(environment_file, filelist, work_queue_executor, flatten=False, compression=2)
        template_analysis(environment_file, filelist, work_queue_executor, flatten=True, compression=2)

def create_conda_environment(env_file, py_version):
    """ Generate a conda environment file 'env_file' to send along the tasks. """

    if os.path.exists(env_file):
        print("conda environment file '{}' already exists. Not generating again.".format(env_file))
        return

    print("creating conda environment file '{}'...".format(env_file))
    import subprocess
    import tempfile

    with tempfile.NamedTemporaryFile(mode='w') as conda_recipe, tempfile.TemporaryDirectory() as tmp_env:
        conda_recipe.write("""
#! /bin/bash
set -e
conda create -y --prefix {tmp_env} python={py_version} conda six dill
source {tmp_env}/bin/activate
conda install -y -c conda-forge xrootd conda-pack
pip install ..
python -c 'import conda_pack; conda_pack.pack(prefix="{tmp_env}", output="{env_file}")'
""".format(env_file=env_file, tmp_env=tmp_env, py_version=py_version))

        conda_recipe.flush()

        subprocess.check_call(['/bin/bash', conda_recipe.name])
        print("done creating conda environment '{}'".format(env_file))


if __name__ == '__main__':
    py_version = "{}.{}".format(sys.version_info[0], sys.version_info[1])  # 3.6 or 3.7, or etc.

    environment_file = 'conda-coffea-wq-env-py{}.tar.gz'.format(py_version)

    create_conda_environment(environment_file, py_version)

    work_queue_example(environment_file)




from __future__ import print_function
import sys

from coffea import lookup_tools
import uproot
from coffea.util import awkward
from coffea.util import numpy as np

import pytest

from dummy_distributions import dummy_jagged_eta_pt, dummy_four_momenta


def jetmet_evaluator():
    from coffea.lookup_tools import extractor
    extract = extractor()

    extract.add_weight_sets(['* * tests/samples/Summer16_23Sep2016V3_MC_L1FastJet_AK4PFPuppi.jec.txt.gz',
                             '* * tests/samples/Summer16_23Sep2016V3_MC_L2L3Residual_AK4PFPuppi.jec.txt.gz',
                             '* * tests/samples/Summer16_23Sep2016V3_MC_L2Relative_AK4PFPuppi.jec.txt.gz',
                             '* * tests/samples/Summer16_23Sep2016V3_MC_L3Absolute_AK4PFPuppi.jec.txt.gz',
                             '* * tests/samples/Summer16_23Sep2016V3_MC_UncertaintySources_AK4PFPuppi.junc.txt.gz',
                             '* * tests/samples/Summer16_23Sep2016V3_MC_Uncertainty_AK4PFPuppi.junc.txt.gz',
                             '* * tests/samples/Fall17_17Nov2017_V6_MC_UncertaintySources_AK4PFchs.junc.txt.gz',
                             '* * tests/samples/Regrouped_Fall17_17Nov2017_V32_MC_UncertaintySources_AK4PFchs.junc.txt',
                             '* * tests/samples/Spring16_25nsV10_MC_PtResolution_AK4PFPuppi.jr.txt.gz',
                             '* * tests/samples/Spring16_25nsV10_MC_SF_AK4PFPuppi.jersf.txt.gz',
                             '* * tests/samples/Autumn18_V7_MC_SF_AK4PFchs.jersf.txt.gz'])

    extract.finalize()

    return extract.make_evaluator()

evaluator = jetmet_evaluator()


def test_factorized_jet_corrector():
    from coffea.jetmet_tools import FactorizedJetCorrector

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    test_Rho = np.full_like(test_eta, 100.)
    test_A = np.full_like(test_eta, 5.)

    jec_names = ['Summer16_23Sep2016V3_MC_L1FastJet_AK4PFPuppi',
                 'Summer16_23Sep2016V3_MC_L2Relative_AK4PFPuppi',
                 'Summer16_23Sep2016V3_MC_L2L3Residual_AK4PFPuppi',
                 'Summer16_23Sep2016V3_MC_L3Absolute_AK4PFPuppi']
    corrector = FactorizedJetCorrector(**{name: evaluator[name] for name in jec_names})

    print(corrector)

    pt_copy = np.copy(test_pt)

    corrs = corrector.getCorrection(JetEta=test_eta, Rho=test_Rho, JetPt=test_pt, JetA=test_A)

    assert((np.abs(pt_copy - test_pt) < 1e-6).all())


def test_jet_resolution():
    from coffea.jetmet_tools import JetResolution

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    test_Rho = np.full_like(test_eta, 100.)

    jer_names = ['Spring16_25nsV10_MC_PtResolution_AK4PFPuppi']
    reso = JetResolution(**{name: evaluator[name] for name in jer_names})

    print(reso)

    resos = reso.getResolution(JetEta=test_eta, Rho=test_Rho, JetPt=test_pt)


def test_jet_correction_uncertainty():
    from coffea.jetmet_tools import JetCorrectionUncertainty

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    junc_names = ['Summer16_23Sep2016V3_MC_Uncertainty_AK4PFPuppi']
    junc = JetCorrectionUncertainty(**{name: evaluator[name] for name in junc_names})

    print(junc)

    juncs = junc.getUncertainty(JetEta=test_eta, JetPt=test_pt)

    for level, corrs in juncs:
        assert(corrs.shape[0] == test_eta.shape[0])


def test_jet_correction_uncertainty_sources():
    from coffea.jetmet_tools import JetCorrectionUncertainty

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    junc_names = []
    levels = []
    for name in dir(evaluator):
        if 'Summer16_23Sep2016V3_MC_UncertaintySources_AK4PFPuppi' in name:
            junc_names.append(name)
            levels.append(name.split('_')[-1])
    junc = JetCorrectionUncertainty(**{name: evaluator[name] for name in junc_names})

    print(junc)

    juncs = junc.getUncertainty(JetEta=test_eta, JetPt=test_pt)

    for level, corrs in juncs:
        assert(level in levels)
        assert(corrs.shape[0] == test_eta.shape[0])


def test_jet_correction_regrouped_uncertainty_sources():
    from coffea.jetmet_tools import JetCorrectionUncertainty

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    junc_names = []
    levels = []
    for name in dir(evaluator):
        if 'Regrouped_Fall17_17Nov2017_V32_MC_UncertaintySources_AK4PFchs' in name:
            junc_names.append(name)
            if len(name.split('_')) == 9:
                levels.append("_".join(name.split('_')[-2:]))
            else:
                levels.append(name.split('_')[-1])
    junc = JetCorrectionUncertainty(**{name: evaluator[name] for name in junc_names})

    print(junc)

    for tpl in list(junc.getUncertainty(JetEta=test_eta, JetPt=test_pt)):
        assert(tpl[0] in levels)
        assert(tpl[1].shape[0] == test_eta.shape[0])

def test_jet_resolution_sf():
    from coffea.jetmet_tools import JetResolutionScaleFactor

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    jersf_names = ['Spring16_25nsV10_MC_SF_AK4PFPuppi']
    resosf = JetResolutionScaleFactor(**{name: evaluator[name] for name in jersf_names})

    print(resosf)

    resosfs = resosf.getScaleFactor(JetEta=test_eta)

def test_jet_resolution_sf_2d():
    from coffea.jetmet_tools import JetResolutionScaleFactor
    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    resosf = JetResolutionScaleFactor(**{name: evaluator[name] for name in ["Autumn18_V7_MC_SF_AK4PFchs"]})
    resosfs = resosf.getScaleFactor(JetPt=test_pt, JetEta=test_eta)

def test_jet_transformer():
    import numpy as np
    import awkward as ak
    import math
    from coffea.analysis_objects import JaggedCandidateArray as CandArray
    from coffea.jetmet_tools import (FactorizedJetCorrector,
                                     JetResolution,
                                     JetResolutionScaleFactor,
                                     JetCorrectionUncertainty,
                                     JetTransformer)

    counts, test_px, test_py, test_pz, test_e = dummy_four_momenta()

    test_Rho = np.full(shape=(np.sum(counts),), fill_value=100.)
    test_A = np.full(shape=(np.sum(counts),), fill_value=5.)

    jets = CandArray.candidatesfromcounts(counts, px=test_px, py=test_py, pz=test_pz, energy=test_e)
    jets.add_attributes(ptRaw=jets.pt,
                        massRaw=jets.mass,
                        rho=test_Rho,
                        area=test_A)

    fakemet = np.random.exponential(scale=1.0,size=counts.size)
    metphi = np.random.uniform(low=-math.pi, high=math.pi, size=counts.size)
    syst_up = 0.001*fakemet
    syst_down = -0.001*fakemet
    met = CandArray.candidatesfromcounts(np.ones_like(counts),
                                         pt=fakemet,
                                         eta=np.zeros_like(counts),
                                         phi=metphi,
                                         mass=np.zeros_like(counts),
                                         MetUnclustEnUpDeltaX=syst_up*np.cos(metphi),
                                         MetUnclustEnUpDeltaY=syst_down*np.sin(metphi))

    jec_names = ['Summer16_23Sep2016V3_MC_L1FastJet_AK4PFPuppi',
                 'Summer16_23Sep2016V3_MC_L2Relative_AK4PFPuppi',
                 'Summer16_23Sep2016V3_MC_L2L3Residual_AK4PFPuppi',
                 'Summer16_23Sep2016V3_MC_L3Absolute_AK4PFPuppi']
    corrector = FactorizedJetCorrector(**{name: evaluator[name] for name in jec_names})

    junc_names = []
    for name in dir(evaluator):
        if 'Summer16_23Sep2016V3_MC_UncertaintySources_AK4PFPuppi' in name:
            junc_names.append(name)
    junc = JetCorrectionUncertainty(**{name: evaluator[name] for name in junc_names})

    jer_names = ['Spring16_25nsV10_MC_PtResolution_AK4PFPuppi']
    reso = JetResolution(**{name: evaluator[name] for name in jer_names})

    jersf_names = ['Spring16_25nsV10_MC_SF_AK4PFPuppi']
    resosf = JetResolutionScaleFactor(**{name: evaluator[name] for name in jersf_names})

    xform = JetTransformer(jec=corrector, junc=junc, jer=reso, jersf=resosf)

    print(xform.uncertainties)

    xform.transform(jets, met=met)

    print('jets',jets.columns)
    print('met',met.columns)

    assert('pt_jer_up' in jets.columns)
    assert('pt_jer_down' in jets.columns)
    assert('mass_jer_up' in jets.columns)
    assert('mass_jer_down' in jets.columns)

    assert('pt_UnclustEn_up' in met.columns)
    assert('pt_UnclustEn_down' in met.columns)
    assert('phi_UnclustEn_up' in met.columns)
    assert('phi_UnclustEn_down' in met.columns)

    for unc in xform.uncertainties:
        assert('pt_'+unc+'_up' in jets.columns)
        assert('pt_'+unc+'_down' in jets.columns)
        assert('mass_'+unc+'_up' in jets.columns)
        assert('mass_'+unc+'_down' in jets.columns)
        assert('pt_'+unc+'_up' in met.columns)
        assert('phi_'+unc+'_up' in met.columns)

def test_jet_correction_uncertainty_sources():
    from coffea.jetmet_tools import JetCorrectionUncertainty

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    junc_names = []
    levels = []
    for name in dir(evaluator):
        if 'Summer16_23Sep2016V3_MC_UncertaintySources_AK4PFPuppi' in name:
            junc_names.append(name)
            levels.append(name.split('_')[-1])
        #test for underscore in dataera
        if 'Fall17_17Nov2017_V6_MC_UncertaintySources_AK4PFchs_AbsoluteFlavMap' in name:
            junc_names.append(name)
            levels.append(name.split('_')[-1])
    junc = JetCorrectionUncertainty(**{name: evaluator[name] for name in junc_names})

    print(junc)

    juncs = junc.getUncertainty(JetEta=test_eta, JetPt=test_pt)

    for level, corrs in juncs:
        assert(level in levels)
        assert(corrs.shape[0] == test_eta.shape[0])



"""Mixin methods for NanoEvents"""
from coffea.nanoevents.methods.base import NanoCollection, NanoEvents
from coffea.nanoevents.methods.vector import (
    TwoVector,
    PolarTwoVector,
    ThreeVector,
    SphericalThreeVector,
    LorentzVector,
    PtEtaPhiMLorentzVector,
)
from coffea.nanoevents.methods.candidate import (
    Candidate,
    PtEtaPhiMCandidate,
)
from coffea.nanoevents.methods.generator import GenParticle, GenVisTau
from coffea.nanoevents.methods.lepton import (
    Electron,
    Muon,
    Tau,
    Photon,
)
from coffea.nanoevents.methods.jetmet import Jet, FatJet, MissingET
from coffea.nanoevents.methods.generic import PtEtaPhiMCollection



import awkward1


def mixin_class(cls):
    """Decorator to register a mixin class

    Fills in any known behaviors based on class inheritance
    """
    name = cls.__name__
    awkward1.behavior[name] = type(name + "Record", (cls, awkward1.Record), {})
    awkward1.behavior["*", name] = type(name + "Array", (cls, awkward1.Array), {})
    for basecls in cls.__mro__:
        for method in basecls.__dict__.values():
            if hasattr(method, "_awkward_mixin"):
                ufunc, rhs, transpose = method._awkward_mixin
                if rhs is None:
                    awkward1.behavior.setdefault((ufunc, name), method)
                    continue
                for rhs_name in list(rhs) + [name]:
                    awkward1.behavior.setdefault((ufunc, name, rhs_name), method)
                    if transpose is not None:
                        awkward1.behavior.setdefault((ufunc, rhs_name, name), transpose)
                if basecls.__name__ in rhs:
                    rhs.add(name)
    return cls


def mixin_method(ufunc, rhs=None, transpose=True):
    """Decorator to register a mixin class method

    Using this decorator ensures that derived classes that are declared
    with the `mixin_class` decorator will also have the behaviors that this
    class has.

    ufunc : numpy.ufunc
        A universal function (or NEP18 callable) that is hooked in awkward1,
        i.e. it can be the first argument of a behavior
    rhs : Set[type] or None
        List of right-hand side argument types (leave None if unary function)
        The left-hand side is expected to always be ``self`` of the parent class
        If the function is not unary or binary, call for help :)
    transpose : bool
        Autmatically create a transpose signature (only makes sense for binary ufuncs)
    """

    def register(method):
        if not isinstance(rhs, (set, type(None))):
            raise ValueError("Expected a set of right-hand-side argument types")
        if transpose and rhs is not None:

            def transposed(left, right):
                return method(right, left)

            method._awkward_mixin = (ufunc, rhs, transposed)
        else:
            method._awkward_mixin = (ufunc, rhs, None)
        return method

    return register



import awkward
from .common import Candidate


class LeptonCommon(Candidate):
    '''Embeds all common gen particle and jet cross-references'''
    def _finalize(self, name, events):
        if 'Jet' in events.columns:
            jets = events['Jet']
            reftype = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(jets.type.to.to))
            reftype.check = False
            embedded_jet = type(jets)(
                self._lazy_crossref,
                args=(self._getcolumn('jetIdx'), jets),
                type=reftype,
            )
            embedded_jet.__doc__ = jets.__doc__
            self['matched_jet'] = embedded_jet
            del self['jetIdx']

        if 'GenPart' in events.columns:
            genpart = events['GenPart']
            reftype = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(genpart.type.to.to))
            reftype.check = False
            embedded_genpart = type(genpart)(
                self._lazy_crossref,
                args=(self._getcolumn('genPartIdx'), genpart),
                type=reftype,
            )
            embedded_genpart.__doc__ = genpart.__doc__
            self['matched_gen'] = embedded_genpart
            del self['genPartIdx']

        # disable this type check due to cyclic reference through jets
        self.type.check = False


class Electron(LeptonCommon):
    '''NanoAOD electron object'''
    FAIL = 0
    'cutBased selection minimum value'
    VETO = 1
    'cutBased selection minimum value'
    LOOSE = 2
    'cutBased selection minimum value'
    MEDIUM = 3
    'cutBased selection minimum value'
    TIGHT = 4
    'cutBased selection minimum value'

    def _finalize(self, name, events):
        super(Electron, self)._finalize(name, events)
        if 'Photon' in events.columns:
            photons = events['Photon']
            reftype = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(photons.type.to.to))
            reftype.check = False
            embedded_photon = type(photons)(
                self._lazy_crossref,
                args=(self._getcolumn('photonIdx'), photons),
                type=reftype,
            )
            embedded_photon.__doc__ = photons.__doc__
            self['matched_photon'] = embedded_photon
            del self['photonIdx']

    @property
    def isVeto(self):
        '''Returns a boolean array marking veto-level cut-based photons'''
        return (self.cutBased & (1 << self.VETO)).astype(bool)

    @property
    def isLoose(self):
        '''Returns a boolean array marking loose cut-based photons'''
        return (self.cutBased & (1 << self.LOOSE)).astype(bool)

    @property
    def isMedium(self):
        '''Returns a boolean array marking medium cut-based photons'''
        return (self.cutBased & (1 << self.MEDIUM)).astype(bool)

    @property
    def isTight(self):
        '''Returns a boolean array marking tight cut-based photons'''
        return (self.cutBased & (1 << self.TIGHT)).astype(bool)


class Muon(LeptonCommon):
    '''NanoAOD muon object'''
    def _finalize(self, name, events):
        super(Muon, self)._finalize(name, events)


class Photon(LeptonCommon):
    '''NanoAOD photon object'''
    LOOSE = 0
    'cutBasedBitmap bit position'
    MEDIUM = 1
    'cutBasedBitmap bit position'
    TIGHT = 2
    'cutBasedBitmap bit position'

    @property
    def isLoose(self):
        '''Returns a boolean array marking loose cut-based photons'''
        return (self.cutBasedBitmap & (1 << self.LOOSE)).astype(bool)

    @property
    def isMedium(self):
        '''Returns a boolean array marking medium cut-based photons'''
        return (self.cutBasedBitmap & (1 << self.MEDIUM)).astype(bool)

    @property
    def isTight(self):
        '''Returns a boolean array marking tight cut-based photons'''
        return (self.cutBasedBitmap & (1 << self.TIGHT)).astype(bool)

    def _finalize(self, name, events):
        del self['mass']
        super(Photon, self)._finalize(name, events)
        if 'Electron' in events.columns:
            electrons = events['Electron']
            reftype = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(electrons.type.to.to))
            reftype.check = False
            embedded_electron = type(electrons)(
                self._lazy_crossref,
                args=(self._getcolumn('electronIdx'), electrons),
                type=reftype,
            )
            embedded_electron.__doc__ = electrons.__doc__
            self['matched_electron'] = embedded_electron
            del self['electronIdx']


class Tau(LeptonCommon):
    '''NanoAOD tau object'''
    def _finalize(self, name, events):
        super(Tau, self)._finalize(name, events)



import awkward1
from coffea.nanoevents.methods.util import apply_global_index
from coffea.nanoevents.methods.mixin import mixin_class, mixin_method
from coffea.nanoevents.methods.base import NanoCollection
from coffea.nanoevents.methods.candidate import PtEtaPhiMCandidate


class CommonMatched:
    @property
    def matched_gen(self):
        return apply_global_index(self.genPartIdxG, self._events().GenPart)

    @property
    def matched_jet(self):
        return apply_global_index(self.jetIdxG, self._events().Jet)


@mixin_class
class Electron(PtEtaPhiMCandidate, NanoCollection, CommonMatched):
    """NanoAOD electron object"""

    FAIL = 0
    "cutBased selection minimum value"
    VETO = 1
    "cutBased selection minimum value"
    LOOSE = 2
    "cutBased selection minimum value"
    MEDIUM = 3
    "cutBased selection minimum value"
    TIGHT = 4
    "cutBased selection minimum value"
    pass

    @property
    def isVeto(self):
        """Returns a boolean array marking veto cut-based photons"""
        return (self.cutBased & (1 << self.VETO)) != 0

    @property
    def isLoose(self):
        """Returns a boolean array marking loose cut-based photons"""
        return (self.cutBased & (1 << self.LOOSE)) != 0

    @property
    def isMedium(self):
        """Returns a boolean array marking medium cut-based photons"""
        return (self.cutBased & (1 << self.MEDIUM)) != 0

    @property
    def isTight(self):
        """Returns a boolean array marking tight cut-based photons"""
        return (self.cutBased & (1 << self.TIGHT)) != 0

    @property
    def matched_photon(self):
        return apply_global_index(self.photonIdxG, self._events().Photon)


@mixin_class
class Muon(PtEtaPhiMCandidate, NanoCollection, CommonMatched):
    """NanoAOD muon object"""


@mixin_class
class Tau(PtEtaPhiMCandidate, NanoCollection, CommonMatched):
    """NanoAOD tau object"""


@mixin_class
class Photon(PtEtaPhiMCandidate, NanoCollection, CommonMatched):
    """NanoAOD photon object"""

    LOOSE = 0
    "cutBasedBitmap bit position"
    MEDIUM = 1
    "cutBasedBitmap bit position"
    TIGHT = 2
    "cutBasedBitmap bit position"

    @property
    def mass(self):
        return awkward1.broadcast_arrays(self.pt, 0.0)[1]

    @property
    def isLoose(self):
        """Returns a boolean array marking loose cut-based photons"""
        return (self.cutBasedBitmap & (1 << self.LOOSE)) != 0

    @property
    def isMedium(self):
        """Returns a boolean array marking medium cut-based photons"""
        return (self.cutBasedBitmap & (1 << self.MEDIUM)) != 0

    @property
    def isTight(self):
        """Returns a boolean array marking tight cut-based photons"""
        return (self.cutBasedBitmap & (1 << self.TIGHT)) != 0

    @property
    def matched_electron(self):
        return apply_global_index(self.electronIdxG, self._events().Electron)



import os
from coffea.nanoaod import NanoEvents


def genroundtrips(genpart):
    # check genpart roundtrip
    assert (genpart.children.parent.pdgId == genpart.pdgId).all().all().all()
    assert (genpart.parent.children.pdgId == genpart.pdgId).any().fillna(True).all().all()
    # distinctParent should be distinct and it should have a relevant child
    assert (genpart.distinctParent.pdgId != genpart.pdgId).fillna(True).all().all()
    assert (genpart.distinctParent.children.pdgId == genpart.pdgId).any().fillna(True).all().all()
    # exercise hasFlags
    genpart.hasFlags(['isHardProcess'])
    genpart.hasFlags(['isHardProcess', 'isDecayedLeptonHadron'])


def crossref(events):
    # check some cross-ref roundtrips (some may not be true always but they are for the test file)
    assert (events.Jet.matched_muons.matched_jet.pt == events.Jet.pt).all().all().all()
    assert (events.Electron.matched_photon.matched_electron.r9 == events.Electron.r9).all().all()


def test_read_nanomc():
    events = NanoEvents.from_file(os.path.abspath('tests/samples/nano_dy.root'))

    # test after views first
    genroundtrips(events.GenPart[events.GenPart.eta > 0])
    genroundtrips(events[(events.Electron.eta > 0).any()].GenPart)
    genroundtrips(events.GenPart)

    # sane gen matching (note for electrons gen match may be photon(22))
    assert ((abs(events.Electron.matched_gen.pdgId) == 11) | (events.Electron.matched_gen.pdgId == 22)).all().all()
    assert (abs(events.Muon.matched_gen.pdgId) == 13).all().all()

    genroundtrips(events.Electron.matched_gen)

    crossref(events[events.Jet.counts > 2])
    crossref(events)

    assert events.Photon.isTight.any().tolist()[:9] == [False, True, True, True, False, False, False, False, False]


def test_read_nanodata():
    events = NanoEvents.from_file(os.path.abspath('tests/samples/nano_dimuon.root'))

    crossref(events)
    crossref(events[events.Jet.counts > 2])



from ..lookup_tools.jersf_lookup import jersf_lookup
import warnings
import re
from ..util import awkward
from ..util import numpy as np
from copy import deepcopy


def _checkConsistency(against, tocheck):
    if against is None:
        against = tocheck
    else:
        if against != tocheck:
            raise Exception('Corrector for {} is mixed'
                            'with correctors for {}!'.format(tocheck, against))
    return tocheck


_levelre = re.compile('SF+')


def _getLevel(levelName):
    matches = _levelre.findall(levelName)
    if len(matches) > 1:
        raise Exception('Malformed JERSF level name: {}'.format(levelName))
    return matches[0]


_level_order = ['SF']


class JetResolutionScaleFactor(object):
    """
    This class is a columnar implementation of the JetResolutionScaleFactor tool in
    CMSSW and FWLite. It calculates the jet energy resolution scale factor for a
    corrected jet in a given binning.

    It implements the jet energy scale factor definition specified in the JER TWiki_.

    .. _TWiki: https://twiki.cern.ch/twiki/bin/view/CMS/JetResolution

    You can use this class as follows::

        jersf = JetResolutionScaleFactor(name1=corrL1,...)
        jetResSF = jersf(JetParameter1=jet.parameter1,...)

    """
    def __init__(self, **kwargs):
        """
        You construct a JetResolutionScaleFactor by passing in a dict of names and functions.
        Names must be formatted as '<campaign>_<dataera>_<datatype>_<level>_<jettype>'.
        """
        jettype = None
        levels = []
        funcs = []
        datatype = None
        campaign = None
        dataera = None
        for name, func in kwargs.items():
            if not isinstance(func, jersf_lookup):
                raise Exception('{} is a {} and not a jersf_lookup!'.format(name, type(func)))
            info = name.split('_')
            if len(info) != 5:
                raise Exception('Corrector name is not properly formatted!')

            campaign = _checkConsistency(campaign, info[0])
            dataera = _checkConsistency(dataera, info[1])
            datatype = _checkConsistency(datatype, info[2])
            levels.append(info[3])
            funcs.append(func)
            jettype = _checkConsistency(jettype, info[4])

        if campaign is None:
            raise Exception('Unable to determine production campaign of JECs!')
        else:
            self._campaign = campaign

        if dataera is None:
            raise Exception('Unable to determine data era of JECs!')
        else:
            self._dataera = dataera

        if datatype is None:
            raise Exception('Unable to determine if JECs are for MC or Data!')
        else:
            self._datatype = datatype

        if len(levels) == 0:
            raise Exception('No levels provided?')
        else:
            self._levels = levels
            self._funcs = funcs

        if jettype is None:
            raise Exception('Unable to determine type of jet to correct!')
        else:
            self._jettype = jettype

        for i, level in enumerate(self._levels):
            this_level = _getLevel(level)
            ord_idx = _level_order.index(this_level)
            if i != this_level:
                self._levels[i], self._levels[ord_idx] = self._levels[ord_idx], self._levels[i]
                self._funcs[i], self._funcs[ord_idx] = self._funcs[ord_idx], self._funcs[i]

        # now we setup the call signature for this factorized JEC
        self._signature = []
        for func in self._funcs:
            sig = func.signature
            for input in sig:
                if input not in self._signature:
                    self._signature.append(input)

    @property
    def signature(self):
        """ list the necessary jet properties that must be input to this function """
        return self._signature

    def __repr__(self):
        out = 'campaign   : %s\n' % (self._campaign)
        out += 'data era   : %s\n' % (self._dataera)
        out += 'data type  : %s\n' % (self._datatype)
        out += 'jet type   : %s\n' % (self._jettype)
        out += 'levels     : %s\n' % (','.join(self._levels))
        out += 'signature  : (%s)\n' % (','.join(self._signature))
        return out

    def getScaleFactor(self, **kwargs):
        """
        Returns the set of resolutions for all input jets at the highest available level

        Use it like::

            jersfs = jersf.getScaleFactor(JetProperty1=jet.property1,...)

        """
        sfs = []
        for i, func in enumerate(self._funcs):
            sig = func.signature
            args = []
            for input in sig:
                args.append(kwargs[input])
            sfs.append(func(*tuple(args)))
        return sfs[-1]



import numpy
import awkward1
from coffea.nanoevents.methods.mixin import mixin_class, mixin_method
from coffea.nanoevents.methods.vector import LorentzVector, PtEtaPhiMLorentzVector


@mixin_class
class Candidate(LorentzVector):
    """A Lorentz vector with charge

    Properties this class requires: x, y, z, t, charge
    """

    @mixin_method(numpy.add, {"Candidate"})
    def add(self, other):
        return awkward1.zip(
            {
                "x": self.x + other.x,
                "y": self.y + other.y,
                "z": self.z + other.z,
                "t": self.t + other.t,
                "charge": self.charge + other.charge,
            },
            with_name="Candidate",
        )

    def sum(self, axis=-1):
        return awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
                "z": awkward1.sum(self.z, axis=axis),
                "t": awkward1.sum(self.t, axis=axis),
                "charge": awkward1.sum(self.charge, axis=axis),
            },
            with_name="Candidate",
        )


@mixin_class
class PtEtaPhiMCandidate(Candidate, PtEtaPhiMLorentzVector):
    """A Lorentz vector in eta, mass coordinates with charge

    Properties this class requires: pt, eta, phi, mass, charge
    """

    pass



import os
import numpy as np
# crystalball is single sided, local reimplementation of double-sided here until
# the PR can be merged
# from scipy.stats import crystalball
from .doublecrystalball import doublecrystalball
from awkward import JaggedArray
from coffea.lookup_tools.dense_lookup import dense_lookup


class rochester_lookup():

    def __init__(self, wrapped_values):
        self._build(wrapped_values)

    def _build(self, wrapped_values):

        self._nsets = wrapped_values['nsets']
        self._members = wrapped_values['members']

        # now build the lookup tables
        # for data scale, simple, just M A in bins of eta,phi
        edges = wrapped_values['edges']['scales']
        M = wrapped_values['values']['M']
        A = wrapped_values['values']['A']
        self._M = {s: {m: {t: dense_lookup(M[s][m][t], edges) for t in M[s][m]} for m in M[s]} for s in M}
        self._A = {s: {m: {t: dense_lookup(A[s][m][t], edges) for t in A[s][m]} for m in A[s]} for s in A}

        # for mc scale, more complicated
        # version 1 if gen pt available
        # only requires the kRes lookup
        edges = wrapped_values['edges']['res']
        kRes = wrapped_values['values']['kRes']
        self._kRes = {s: {m: {t: dense_lookup(kRes[s][m][t], edges) for t in kRes[s][m]} for m in kRes[s]} for s in kRes}

        # version 2 if gen pt not available
        edges = wrapped_values['edges']['cb']
        rsPars = wrapped_values['values']['rsPars']
        cbS = wrapped_values['values']['cbS']
        cbA = wrapped_values['values']['cbA']
        cbN = wrapped_values['values']['cbN']
        self._rsPars = {s: {m: {t: dense_lookup(rsPars[s][m][t], edges) for t in rsPars[s][m]} for m in rsPars[s]} for s in rsPars}
        self._cbS = {s: {m: dense_lookup(cbS[s][m], edges) for m in cbS[s]} for s in cbS}
        self._cbA = {s: {m: dense_lookup(cbA[s][m], edges) for m in cbA[s]} for s in cbA}
        self._cbN = {s: {m: dense_lookup(cbN[s][m], edges) for m in cbN[s]} for s in cbN}

        self._loaduncs = len(self._M.keys()) > 1

    def _error(self, func, *args):
        if not self._loaduncs:
            return None

        newargs = args + (0, 0)
        default = func(*newargs)
        result = np.zeros_like(default)
        for s in range(self._nsets):
            oneOver = 1.0 / self._members[s]
            for m in range(self._members[s]):
                newargs = args + (s, m)
                d = func(*newargs) - default
                result = result + d * d * oneOver
        return result**0.5

    def kScaleDT(self, charge, pt, eta, phi, s=0, m=0):
        '''Momentum scale correction for data
            required:
                charge
                pt
                eta
                phi
            optional:
                s: Rochester correction set to use (default 0)
                m: Rochester correction member to use (default 0)
        '''
        # type = 1 corresponds to data
        M = self._M[s][m][1](eta, phi)
        A = self._A[s][m][1](eta, phi)
        return 1.0 / (M + charge * A * pt)

    def kScaleDTerror(self, charge, pt, eta, phi):
        '''Momentum scale correction uncertainty for data
            required:
                charge
                pt
                eta
                phi
        '''
        return self._error(self.kScaleDT, charge, pt, eta, phi)

    def kScaleMC(self, charge, pt, eta, phi, s=0, m=0):
        '''Momentum scale correction for mc (not recommended, use kSpreadMC instead)
            required:
                charge
                pt
                eta
                phi
            optional:
                s: Rochester correction set to use (default 0)
                m: Rochester correction member to use (default 0)
        '''
        # type = 0 corresponds to mc
        M = self._M[s][m][0](eta, phi)
        A = self._A[s][m][0](eta, phi)
        return 1.0 / (M + charge * A * pt)

    def kScaleMCerror(self, charge, pt, eta, phi):
        '''Momentum scale correction uncertainty for mc (not recommended, use kSpreadMC instead)
            required:
                charge
                pt
                eta
                phi
        '''
        return self._error(self.kScaleMC, charge, pt, eta, phi)

    def kSpreadMC(self, charge, pt, eta, phi, genpt, s=0, m=0):
        '''Momentum scale correction for mc (if genpt not availble, use kSmearMC)
            required:
                charge
                pt
                eta
                phi
                genpt: the match gen particle pt
            optional:
                s: Rochester correction set to use (default 0)
                m: Rochester correction member to use (default 0)
        '''
        k = self.kScaleMC(charge, pt, eta, phi, s, m)
        return k * self._kSpread(genpt, k * pt, eta, s, m)

    def kSpreadMCerror(self, charge, pt, eta, phi, genpt):
        '''Momentum scale correction uncertainty for mc (if genpt not availble, use kSmearMC)
            required:
                charge
                pt
                eta
                phi
                genpt: the match gen particle pt
        '''
        return self._error(self.kSpreadMC, charge, pt, eta, phi, genpt)

    def _kSpread(self, genpt, kpt, eta, s=0, m=0):
        x = genpt / kpt
        abseta = abs(eta)
        kData = self._kRes[s][m][1](abseta)  # type 1 is data
        kMC = self._kRes[s][m][0](abseta)    # type 0 is MC
        return x / (1.0 + (x - 1.0) * kData / kMC)

    def kSmearMC(self, charge, pt, eta, phi, nl, u, s=0, m=0):
        '''Momentum scale correction for mc (should prefer to use kSpreadMC)
            required:
                charge
                pt
                eta
                phi
                nl: number of tracker layers with measurements
                u: random float between 0 and 1
            optional:
                s: Rochester correction set to use (default 0)
                m: Rochester correction member to use (default 0)
        '''
        k = self.kScaleMC(charge, pt, eta, phi, s, m)
        return k * self._kExtra(k * pt, eta, nl, u, s, m)

    def kSmearMCerror(self, charge, pt, eta, phi, nl, u):
        '''Momentum scale correction uncertainty for mc (should prefer to use kSpreadMC)
            required:
                charge
                pt
                eta
                phi
                nl: number of tracker layers with measurements
                u: random float between 0 and 1
        '''
        return self._error(self.kSmearMC, charge, pt, eta, phi, nl, u)

    def _sigma(self, pt, eta, nl, s=0, m=0):
        dpt = pt - 45
        abseta = abs(eta)
        return self._rsPars[s][m][0](abseta, nl) + self._rsPars[s][m][1](abseta, nl) * dpt + self._rsPars[s][m][2](abseta, nl) * dpt * dpt

    def _kExtra(self, kpt, eta, nl, u, s=0, m=0):
        # if it is a jagged array, save the offsets then flatten everything
        # needed for the ternary conditions later
        offsets = None
        if isinstance(kpt, JaggedArray):
            offsets = kpt.offsets
            kpt = kpt.flatten()
            eta = eta.flatten()
            nl = nl.flatten()
            u = u.flatten()
        abseta = abs(eta)
        kData = self._kRes[s][m][1](abseta)  # type 1 is data
        kMC = self._kRes[s][m][0](abseta)    # type 0 is MC
        mask = (kData > kMC)
        x = np.zeros_like(kpt)
        sigma = self._sigma(kpt, eta, nl, s, m)
        # Rochester cbA = beta, cbN = m, as well as cbM (always 0?) = loc and cbS = scale to transform y = (x-loc)/scale in the pdf method
        cbA = self._cbA[s][m](abseta, nl)
        cbN = self._cbN[s][m](abseta, nl)
        loc = np.zeros_like(u)
        cbS = self._cbS[s][m](abseta, nl)
        invcdf = doublecrystalball.ppf(u, cbA, cbA, cbN, cbN, loc, cbS)
        x[mask] = np.sqrt(kData[mask] * kData[mask] - kMC[mask] * kMC[mask]) * sigma[mask] * invcdf[mask]
        result = np.ones_like(kpt)
        result[(x > -1)] = 1.0 / (1.0 + x[x > -1])
        if offsets is not None:
            result = JaggedArray.fromoffsets(offsets, result)
        return result



from coffea.nanoevents.methods.mixin import mixin_class
from coffea.nanoevents.methods.base import NanoCollection
from coffea.nanoevents.methods.vector import PtEtaPhiMLorentzVector


@mixin_class
class PtEtaPhiMCollection(PtEtaPhiMLorentzVector, NanoCollection):
    pass



from __future__ import print_function, division

from coffea.util import numpy as np
from dummy_distributions import dummy_jagged_eta_pt

import sys
import os.path as osp
import pytest

def test_weights():
    from coffea.processor import Weights
    
    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    scale_central = np.random.normal(loc=1.0, scale=0.01, size=counts.size)
    scale_up = scale_central * 1.10
    scale_down = scale_central * 0.95
    scale_up_shift = 0.10 * scale_central
    scale_down_shift = 0.05 * scale_central

    weight = Weights(counts.size)
    weight.add('test', scale_central, weightUp=scale_up, weightDown=scale_down)
    weight.add('testShift', scale_central, weightUp=scale_up_shift,
               weightDown=scale_down_shift, shift=True)

    var_names = weight.variations
    expected_names = ['testShiftUp', 'testShiftDown', 'testUp', 'testDown']
    for name in expected_names:
        assert(name in var_names)

    test_central = weight.weight()
    exp_weight = scale_central * scale_central

    assert(np.all(np.abs(test_central - (exp_weight)) < 1e-6))

    test_up = weight.weight('testUp')
    exp_up = scale_central * scale_central * 1.10

    assert(np.all(np.abs(test_up - (exp_up)) < 1e-6))

    test_down = weight.weight('testDown')
    exp_down = scale_central * scale_central * 0.95

    assert(np.all(np.abs(test_down - (exp_down)) < 1e-6))

    test_shift_up = weight.weight('testUp')

    assert(np.all(np.abs(test_shift_up - (exp_up)) < 1e-6))
    
    test_shift_down = weight.weight('testDown')

    assert(np.all(np.abs(test_shift_down - (exp_down)) < 1e-6))


def test_weights_partial():
    from coffea.processor import Weights
    counts, _, _ = dummy_jagged_eta_pt()
    w1 = np.random.normal(loc=1.0, scale=0.01, size=counts.size)
    w2 = np.random.normal(loc=1.3, scale=0.05, size=counts.size)

    weights = Weights(counts.size, storeIndividual=True)
    weights.add('w1', w1)
    weights.add('w2', w2)

    test_exclude_none = weights.weight()
    assert(np.all(np.abs(test_exclude_none - w1 * w2)<1e-6))

    test_exclude1 = weights.partial_weight(exclude=['w1'])
    assert(np.all(np.abs(test_exclude1 - w2)<1e-6))

    test_include1 = weights.partial_weight(include=['w1'])
    assert(np.all(np.abs(test_include1 - w1)<1e-6))

    test_exclude2 = weights.partial_weight(exclude=['w2'])
    assert(np.all(np.abs(test_exclude2 - w1)<1e-6))

    test_include2 = weights.partial_weight(include=['w2'])
    assert(np.all(np.abs(test_include2 - w2)<1e-6))

    test_include_both = weights.partial_weight(include=['w1','w2'])
    assert(np.all(np.abs(test_include_both - w1 * w2)<1e-6))

    # Check that exception is thrown if arguments are incompatible
    error_raised = False
    try:
        weights.partial_weight(exclude=['w1'], include=['w2'])
    except ValueError:
        error_raised = True
    assert(error_raised)

    error_raised = False
    try:
        weights.partial_weight()
    except ValueError:
        error_raised = True
    assert(error_raised)

    # Check that exception is thrown if individual weights
    # are not saved from the start
    weights = Weights(counts.size, storeIndividual=False)
    weights.add('w1', w1)
    weights.add('w2', w2)

    error_raised = False
    try:
        weights.partial_weight(exclude=['test'], include=['test'])
    except ValueError:
        error_raised = True
    assert(error_raised)

def test_packed_selection():
    from coffea.processor import PackedSelection

    sel = PackedSelection()

    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    all_true = np.full(shape=counts.shape, fill_value=True, dtype=np.bool)
    all_false = np.full(shape=counts.shape, fill_value=False, dtype=np.bool)
    ones = np.ones(shape=counts.shape, dtype=np.uint64)
    wrong_shape = ones = np.ones(shape=(counts.shape[0]-5,), dtype=np.bool)

    sel.add('all_true', all_true)
    sel.add('all_false', all_false)

    assert(np.all(sel.require(all_true=True, all_false=False) == all_true))
    assert(np.all(sel.all('all_true', 'all_false') == all_false))

    try:
        sel.require(all_true=1, all_false=0)
    except ValueError:
        pass

    try:
        sel.add('wrong_shape',wrong_shape)
    except ValueError:
        pass

    try:
        sel.add('ones',ones)
    except ValueError:
        pass

    try:
        overpack = PackedSelection()
        for i in range(65):
            overpack.add('sel_%d',all_true)
    except RuntimeError:
        pass

def test_processorabc():
    from coffea.processor import ProcessorABC

    class test(ProcessorABC):
        @property
        def accumulator(self):
            pass
    
        def process(self, df):
            pass
    
        def postprocess(self, accumulator):
            pass
    
    try:
        proc = ProcessorABC()
    except TypeError:
        pass

    proc = test()

    super(test, proc).accumulator

    df = None
    super(test, proc).process(df)

    acc = None
    super(test, proc).postprocess(acc)


@pytest.mark.skipif(sys.platform.startswith("win"), reason='problems with paths on windows')
def test_lazy_dataframe():
    import uproot
    from coffea.processor import LazyDataFrame
    
    tree = uproot.open(osp.abspath('tests/samples/nano_dy.root'))['Events']
    entrystart = 0
    entrystop = 100
    
    df = LazyDataFrame(tree, entrystart, entrystop, preload_items = ['nMuon'])

    assert(len(df) == 1)
    
    pt = df['Muon_pt']
    assert(len(df) == 2)
    df['Muon_pt_up'] = pt * 1.05
    assert(len(df) == 3)
    assert('Muon_pt' in df.materialized)
    
    assert('Muon_eta' in df.available)
    
    assert(df.size == tree.numentries)

    with pytest.raises(KeyError):
        x = df['notthere']


@pytest.mark.skipif(sys.platform.startswith("win"), reason='problems with paths on windows')
def test_lazy_dataframe_getattr():
    import uproot
    from coffea.processor import LazyDataFrame
    
    tree = uproot.open(osp.abspath('tests/samples/nano_dy.root'))['Events']
    entrystart = 0
    entrystop = 100
    
    df = LazyDataFrame(tree, entrystart, entrystop, preload_items = ['nMuon'])

    assert(len(df) == 1)
    
    pt = df.Muon_pt
    assert(len(df) == 2)
    assert('Muon_pt' in df.materialized)
    
    assert('Muon_eta' in df.available)
    
    assert(df.size == tree.numentries)

    with pytest.raises(AttributeError):
        x = df.notthere

    import copy
    df2 = copy.copy(df)
    pt = df2.Muon_pt
    with pytest.raises(AttributeError):
        df2.notthere


@pytest.mark.skipif(sys.platform.startswith("win"), reason='problems with paths on windows')
def test_preloaded_dataframe():
    import uproot
    from coffea.processor import PreloadedDataFrame

    tree = uproot.open(osp.abspath('tests/samples/nano_dy.root'))['Events']
    chunksize = 20
    index = 0
    
    arrays = tree.arrays()
    
    df = PreloadedDataFrame(arrays[b'nMuon'].size, arrays)

    assert(len(arrays) == len(df))
    
    df['nMuon'] = arrays[b'nMuon']
    assert('nMuon' in df.available)
    
    assert(np.all(df[b'nMuon'] == arrays[b'nMuon']))

    assert(b'Muon_eta' in df.available)
    assert(b'nMuon' in df.materialized)
    assert(df.size == arrays[b'nMuon'].size)


@pytest.mark.skipif(sys.platform.startswith("win"), reason='problems with paths on windows')
def test_preloaded_dataframe_getattr():
    import uproot
    from coffea.processor import PreloadedDataFrame

    tree = uproot.open(osp.abspath('tests/samples/nano_dy.root'))['Events']
    chunksize = 20
    index = 0
    
    arrays = tree.arrays()
    
    df = PreloadedDataFrame(arrays[b'nMuon'].size, arrays)

    assert(len(arrays) == len(df))
    
    df['nMuon'] = arrays[b'nMuon']
    assert('nMuon' in df.available)
    
    assert(np.all(df.nMuon == arrays[b'nMuon']))

    assert(b'Muon_eta' in df.available)
    assert('nMuon' in df.materialized)
    assert(df.size == arrays[b'nMuon'].size)



from __future__ import print_function, division
from coffea import processor

import warnings

import numpy as np

import sys
import pytest


def do_dask_job(client, filelist, compression=0):
    treename='Events'
    from coffea.processor.test_items import NanoTestProcessor
    proc = NanoTestProcessor()
    
    exe_args = {
        'client': client,
        'compression': compression,
    }
    hists = processor.run_uproot_job(filelist,
                                     treename,
                                     processor_instance=proc,
                                     executor=processor.dask_executor,
                                     executor_args=exe_args)
    
    assert( hists['cutflow']['ZJets_pt'] == 18 )
    assert( hists['cutflow']['ZJets_mass'] == 6 )
    assert( hists['cutflow']['Data_pt'] == 84 )
    assert( hists['cutflow']['Data_mass'] == 66 )
    

def do_dask_cached(client, filelist, cachestrategy=None):
    from coffea.processor.test_items import NanoEventsProcessor
    from coffea.processor.dask import register_columncache
    register_columncache(client)

    exe_args = {
        'client': client,
        'nano': True,
        'cachestrategy': cachestrategy,
        'savemetrics': True,
        'worker_affinity': True if cachestrategy is not None else False,
    }
    hists, metrics = processor.run_uproot_job(
        filelist,
        'Events',
        processor_instance=NanoEventsProcessor(canaries=['0001a210a3f8364811eaa29ff5b55c90beef;Events;0;40;Muon_pt']),
        executor=processor.dask_executor,
        executor_args=exe_args
    )

    assert( hists['cutflow']['ZJets_pt'] == 18 )
    assert( hists['cutflow']['ZJets_mass'] == 6 )
    assert( hists['cutflow']['Data_pt'] == 84 )
    assert( hists['cutflow']['Data_mass'] == 66 )
    return hists['worker']


def test_dask_local():
    distributed = pytest.importorskip("distributed", minversion="2.6.0")
    # `python setup.py pytest` doesn't seem to play nicely with separate processses
    client = distributed.Client(processes=False, dashboard_address=None)

    import os
    import os.path as osp

    filelist = {
        'ZJets': [osp.join(os.getcwd(),'tests/samples/nano_dy.root')],
        'Data' : [osp.join(os.getcwd(),'tests/samples/nano_dimuon.root')]
    }

    do_dask_job(client, filelist)
    do_dask_job(client, filelist, compression=2)
    do_dask_cached(client, filelist)
    workers1 = do_dask_cached(client, filelist, 'dask-worker')
    assert len(workers1) > 0
    workers2 = do_dask_cached(client, filelist, 'dask-worker')
    assert workers1 == workers2

    filelist = {
        'ZJets': {'treename': 'Events', 'files': [osp.join(os.getcwd(),'tests/samples/nano_dy.root')]},
        'Data' : {'treename': 'Events', 'files': [osp.join(os.getcwd(),'tests/samples/nano_dimuon.root')]}
    }

    do_dask_job(client, filelist)

    client.close()


if __name__ == '__main__':
    test_dask_local()



import os
import awkward1 as ak
from coffea.nanoevents import NanoEventsFactory


def genroundtrips(genpart):
    # check genpart roundtrip
    assert ak.all(genpart.children.parent.pdgId == genpart.pdgId)
    assert ak.all(ak.any(genpart.parent.children.pdgId == genpart.pdgId, axis=-1, mask_identity=True))
    # distinctParent should be distinct and it should have a relevant child
    assert ak.all(genpart.distinctParent.pdgId != genpart.pdgId)
    assert ak.all(ak.any(genpart.distinctParent.children.pdgId == genpart.pdgId, axis=-1, mask_identity=True))
    # exercise hasFlags
    genpart.hasFlags(['isHardProcess'])
    genpart.hasFlags(['isHardProcess', 'isDecayedLeptonHadron'])


def crossref(events):
    # check some cross-ref roundtrips (some may not be true always but they are for the test file)
    assert ak.all(events.Jet.matched_muons.matched_jet.pt == events.Jet.pt)
    assert ak.all(events.Electron.matched_photon.matched_electron.r9 == events.Electron.r9)


def test_read_nanomc():
    factory = NanoEventsFactory(os.path.abspath('tests/samples/nano_dy.root'))
    events = factory.events()

    # test after views first
    genroundtrips(events.GenPart.mask[events.GenPart.eta > 0])
    genroundtrips(events.mask[ak.any(events.Electron.pt > 50, axis=1)].GenPart)
    genroundtrips(events.GenPart)

    genroundtrips(events.GenPart[events.GenPart.eta > 0])
    genroundtrips(events[ak.any(events.Electron.pt > 50, axis=1)].GenPart)

    # sane gen matching (note for electrons gen match may be photon(22))
    assert ak.all((abs(events.Electron.matched_gen.pdgId) == 11) | (events.Electron.matched_gen.pdgId == 22))
    assert ak.all(abs(events.Muon.matched_gen.pdgId) == 13)

    genroundtrips(events.Electron.matched_gen)

    crossref(events[ak.num(events.Jet) > 2])
    crossref(events)

    assert ak.any(events.Photon.isTight, axis=1).tolist()[:9] == [False, True, True, True, False, False, False, False, False]


def test_read_nanodata():
    factory = NanoEventsFactory(os.path.abspath('tests/samples/nano_dimuon.root'))
    events = factory.events()

    crossref(events)
    crossref(events[ak.num(events.Jet) > 2])



from __future__ import print_function, division
from coffea import processor
from functools import partial
import numpy as np


def test_accumulators():
    a = processor.value_accumulator(float)
    a += 3.
    assert a.value == 3.
    assert a.identity().value == 0.

    a = processor.value_accumulator(partial(np.array, [2.]))
    a += 3.
    assert np.array_equal(a.value, np.array([5.]))
    assert np.array_equal(a.identity().value, np.array([2.]))

    l = processor.list_accumulator(range(4))
    l += [3]
    l += processor.list_accumulator([1, 2])
    assert l == [0, 1, 2, 3, 3, 1, 2]

    b = processor.set_accumulator({'apples', 'oranges'})
    b += {'pears'}
    b += 'grapes'
    assert b == {'apples', 'oranges', 'pears', 'grapes'}

    c = processor.dict_accumulator({'num': a, 'fruit': b})
    c['num'] += 2.
    c += processor.dict_accumulator({
        'num2': processor.value_accumulator(int),
        'fruit': processor.set_accumulator({'apples', 'cherries'}),
    })
    assert c['num2'].value == 0
    assert np.array_equal(c['num'].value, np.array([7.]))
    assert c['fruit'] == {'apples', 'oranges', 'pears', 'grapes', 'cherries'}

    d = processor.defaultdict_accumulator(float)
    d['x'] = 0.
    d['x'] += 4.
    d['y'] += 5.
    d['z'] += d['x']
    d['x'] += d['y']
    assert d['x'] == 9.
    assert d['y'] == 5.
    assert d['z'] == 4.
    assert d['w'] == 0.

    e = d + c

    f = processor.defaultdict_accumulator(lambda: 2.)
    f['x'] += 4.
    assert f['x'] == 6.

    f += f
    assert f['x'] == 12.
    assert f['y'] == 2.

    a = processor.column_accumulator(np.arange(6).reshape(2,3))
    b = processor.column_accumulator(np.arange(12).reshape(4,3))
    a += b
    assert a.value.sum() == 81



import numpy as np
from scipy.stats._continuous_distns import rv_continuous, _norm_pdf_C, _norm_cdf, _norm_ppf, sc, _lazywhere


class doublecrystalball_gen(rv_continuous):
    r"""
    Double-sided Crystalball distribution
    %(before_notes)s
    Notes
    -----
    The probability density function for `doublecrystalball` is:
    .. math::
        f(x, \beta_{L}, \beta_{H}, m_{L}, m_{H})= \begin{cases}
                                                      N \exp(-x^2 / 2),  &\text{for } -\beta_{L} < x < \beta_{H}\\
                                                      N A_{L} (B_{L} - x)^{-m_{L}}  &\text{for } x \le -\beta_{L}\\
                                                      N A_{H} (B_{H} - x)^{-m_{H}}  &\text{for } x \ge \beta_{H}\\
                                                    \end{cases}
    where :math:`A_{i}=(m_{i} / |\beta_{i}|)^m_{i}  \exp(-\beta_{i}^2 / 2)`,
    :math:`B_{i}=m_{i}/|\beta_{i}| - |\beta_{i}|` and :math:`N` is a normalisation constant.
    `doublecrystalball` takes :math:`\beta_{i} > 0` and :math:`m_{i} > 1` as shape
    parameters.  :math:`\beta_{i}` defines the point where the pdf changes
    from a power-law to a Gaussian distribution.  :math:`m_{i}` is the power
    of the power-law tail. Here, :math:`i=L, H` refers to the low and high tails, respectively.
    References
    ----------
    .. [1] "Crystal Ball Function",
           https://en.wikipedia.org/wiki/Crystal_Ball_function
    %(after_notes)s
    .. versionadded:: 0.19.0
    %(example)s
    """

    def _pdf(self, x, betaL, betaH, mL, mH):
        """
        Return PDF of the double-sided crystalball function.
        """
        N = 1.0 / (mL / betaL / (mL - 1) * np.exp(-0.5 * betaL * betaL) +
                   mH / betaH / (mH - 1) * np.exp(-0.5 * betaH * betaH) +
                   _norm_pdf_C * (_norm_cdf(betaH) - _norm_cdf(-betaL)))

        def core(x, beta, m):
            return np.exp(-0.5 * x * x)

        def tail(x, beta, m):
            return ((m / beta)**m * np.exp(-0.5 * beta * beta) *
                    (m / beta - beta - x)**(-m))

        def lhs(x, betaL, betaH, mL, mH):
            return tail(x, betaL, mL)

        def rhs(x, betaL, betaH, mL, mH):
            return _lazywhere(x < betaH, (-x, betaH, mH), f=core, f2=tail)

        return N * _lazywhere(x > -betaL, (x, betaL, betaH, mL, mH), f=rhs, f2=lhs)

    def _logpdf(self, x, betaL, betaH, mL, mH):
        """
        Return the log of the PDF of the double-sided crystalball function.
        """
        N = 1.0 / (mL / betaL / (mL - 1) * np.exp(-0.5 * betaL * betaL) +
                   mH / betaH / (mH - 1) * np.exp(-0.5 * betaH * betaH) +
                   _norm_pdf_C * (_norm_cdf(betaH) - _norm_cdf(-betaL)))

        def core(x, beta, m):
            return -0.5 * x * x

        def tail(x, beta, m):
            return m * np.log(m / beta) - 0.5 * beta * beta - m * np.log(m / beta - beta - x)

        def lhs(x, betaL, betaH, mL, mH):
            return tail(x, betaL, mL)

        def rhs(x, betaL, betaH, mL, mH):
            return _lazywhere(x < betaH, (-x, betaH, mH), f=core, f2=tail)

        return np.log(N) + _lazywhere(x > -betaL, (x, betaL, betaH, mL, mH), f=rhs, f2=lhs)

    def _cdf(self, x, betaL, betaH, mL, mH):
        """
        Return CDF of the double-sided crystalball function
        """
        N = 1.0 / (mL / betaL / (mL - 1) * np.exp(-0.5 * betaL * betaL) +
                   mH / betaH / (mH - 1) * np.exp(-0.5 * betaH * betaH) +
                   _norm_pdf_C * (_norm_cdf(betaH) - _norm_cdf(-betaL)))

        def inttail(beta, m):
            return m / beta / (m - 1) * np.exp(-0.5 * beta * beta)

        def intcore(betaL, betaH):
            return _norm_pdf_C * (_norm_cdf(betaH) - _norm_cdf(-betaL))

        def tail(x, beta, m):
            return ((m / beta)**m * np.exp(-0.5 * beta * beta) *
                    (m / beta - beta - x)**(1 - m) / (m - 1))

        def hightail(x, betaL, betaH, mL, mH):
            return (inttail(betaL, mL) + intcore(betaL, betaH) +
                    inttail(betaH, mH) - tail(-x, betaH, mH))

        def core(x, betaL, betaH, mL, mH):
            return (inttail(betaL, mL) + _norm_pdf_C * (_norm_cdf(x) - _norm_cdf(-betaL)))

        def lhs(x, betaL, betaH, mL, mH):
            return tail(x, betaL, mL)

        def rhs(x, betaL, betaH, mL, mH):
            return _lazywhere(x < betaH, (x, betaL, betaH, mL, mH), f=core, f2=hightail)

        return N * _lazywhere(x > -betaL, (x, betaL, betaH, mL, mH), f=rhs, f2=lhs)

    def _ppf(self, p, betaL, betaH, mL, mH):
        """
        Return PPF of the double-sided crystalball function
        """
        def inttail(beta, m):
            return m / beta / (m - 1) * np.exp(-0.5 * beta * beta)

        def intcore(betaL, betaH):
            return _norm_pdf_C * (_norm_cdf(betaH) - _norm_cdf(-betaL))

        def hightail(p, betaL, betaH, mL, mH):
            CL = inttail(betaL, mL)
            CH = inttail(betaH, mH)
            C = CL + CH
            N = 1 / (C + intcore(betaL, betaH))
            eb2H = np.exp(-0.5 * betaH * betaH)
            return -(mH / betaH - betaH -
                     ((mH - 1) * (mH / betaH)**(-mH) / eb2H * (1 - p) / N)**(1 / (1 - mH)))

        def lowtail(p, betaL, betaH, mL, mH):
            CL = inttail(betaL, mL)
            CH = inttail(betaH, mH)
            C = CL + CH
            N = 1 / (C + intcore(betaL, betaH))
            eb2L = np.exp(-0.5 * betaL * betaL)
            return (mL / betaL - betaL -
                    ((mL - 1) * (mL / betaL)**(-mL) / eb2L * p / N)**(1 / (1 - mL)))

        def core(p, betaL, betaH, mL, mH):
            CL = inttail(betaL, mL)
            CH = inttail(betaH, mH)
            C = CL + CH
            N = 1 / (C + intcore(betaL, betaH))
            return _norm_ppf(_norm_cdf(-betaL) + (1 / _norm_pdf_C) * (p / N - CL))

        def ppf_greater(p, betaL, betaH, mL, mH):
            N = 1.0 / (inttail(betaL, mL) + intcore(betaL, betaH) + inttail(betaH, mH))
            pbetaH = 1 - (N * (mH / betaH) * np.exp(-0.5 * betaH * betaH) / (mH - 1))
            return _lazywhere(p > pbetaH, (p, betaL, betaH, mL, mH), f=hightail, f2=core)

        N = 1.0 / (inttail(betaL, mL) + intcore(betaL, betaH) + inttail(betaH, mH))
        pbetaL = N * (mL / betaL) * np.exp(-0.5 * betaL * betaL) / (mL - 1)
        return _lazywhere(p < pbetaL, (p, betaL, betaH, mL, mH), f=lowtail, f2=ppf_greater)

    def _munp(self, n, betaL, betaH, mL, mH):
        """
        Returns the n-th non-central moment of the double-sided crystalball function.
        """
        # this should be copied from crystalball and updated
        raise(NotImplementedError)

    def _argcheck(self, betaL, betaH, mL, mH):
        """
        Shape parameter bounds are m > 1 and beta > 0.
        """
        return (mL > 1) & (betaL > 0) & (mH > 1) & (betaH > 1)


doublecrystalball = doublecrystalball_gen(name='doublecrystalball', longname="A Double-sided Crystalball Function")



"""Physics candidate wrapper classes for awkward arrays
"""
from .JaggedCandidateArray import JaggedCandidateArray
from .JaggedCandidateMethods import JaggedCandidateMethods
from .JaggedTLorentzVectorArray import JaggedTLorentzVectorArray

__all__ = [
    'JaggedCandidateArray',
    'JaggedCandidateMethods',
    'JaggedTLorentzVectorArray'
]



import pytest
import os
import uproot
from coffea import processor
from coffea.processor.test_items import NanoEventsProcessor
from coffea.nanoaod import NanoEvents


def test_preloaded_nanoevents():
    columns = ['nMuon','Muon_pt','Muon_eta','Muon_phi','Muon_mass','Muon_charge', 'nJet', 'Jet_eta']
    p = NanoEventsProcessor(columns=columns)

    tree = uproot.open(os.path.abspath('tests/samples/nano_dy.root'))['Events']
    arrays = tree.arrays(columns, flatten=True, namedecode='ascii')
    df = processor.PreloadedDataFrame(tree.numentries, arrays)
    print(arrays)

    events = NanoEvents.from_arrays(arrays, metadata={'dataset': 'ZJets'})
    hists = p.process(events)

    print(hists)
    assert( hists['cutflow']['ZJets_pt'] == 18 )
    assert( hists['cutflow']['ZJets_mass'] == 6 )

    with pytest.raises(RuntimeError):
        print(events.Muon.matched_jet)



import os
import random
import numpy as np
import uproot
from awkward import JaggedArray

import ROOT
ROOT.PyConfig.IgnoreCommandLineOptions = True

from coffea import lookup_tools

cfname = 'rochester/roccor.Run2.v3/RoccoR.cc'
txtname = 'rochester/roccor.Run2.v3/RoccoR2018.txt'
treename = 'Events'
datafname = 'tests/samples/nano_dimuon.root'
mcfname = 'tests/samples/nano_dy.root'


# official version of rochester
PWD = os.getcwd()
ROOT.gROOT.ProcessLine(f'.L {PWD}/{cfname}')
roccor = ROOT.RoccoR(f'{PWD}/{txtname}')

for isData, fname in [(True,datafname), (False,mcfname)]:
    branches = ['Muon_charge','Muon_pt','Muon_eta','Muon_phi']
    if not isData: branches += ['Muon_genPartIdx', 'GenPart_pt', 'Muon_nTrackerLayers']
    res = []
    err = []
    fullu = []
    for i,arrays in enumerate(uproot.iterate(fname, treename, branches=branches, namedecode='utf-8', entrysteps=200000)):
        charge = arrays['Muon_charge']
        pt = arrays['Muon_pt']
        eta = arrays['Muon_eta']
        phi = arrays['Muon_phi']
        if not isData:
            # for default if gen present
            gid = arrays['Muon_genPartIdx']
            gpt = arrays['GenPart_pt']
            # for backup w/o gen
            nl = arrays['Muon_nTrackerLayers']
            u = np.random.rand(*pt.flatten().shape)
            u = JaggedArray.fromoffsets(pt.offsets, u)
            fullu += [u]
        for ie in range(len(pt)):
            subres = []
            suberr = []
            for im in range(len(pt[ie])):
                if isData:
                    subres += [roccor.kScaleDT(int(charge[ie][im]), float(pt[ie][im]), float(eta[ie][im]), float(phi[ie][im]))]
                    suberr += [roccor.kScaleDTerror(int(charge[ie][im]), float(pt[ie][im]), float(eta[ie][im]), float(phi[ie][im]))]
                else:
                    if gid[ie][im]>=0:
                        subres += [roccor.kSpreadMC(int(charge[ie][im]), float(pt[ie][im]), float(eta[ie][im]), float(phi[ie][im]), float(gpt[ie][gid[ie][im]]))]
                        suberr += [roccor.kSpreadMCerror(int(charge[ie][im]), float(pt[ie][im]), float(eta[ie][im]), float(phi[ie][im]), float(gpt[ie][gid[ie][im]]))]
                    else:
                        subres += [roccor.kSmearMC(int(charge[ie][im]), float(pt[ie][im]), float(eta[ie][im]), float(phi[ie][im]), int(nl[ie][im]), float(u[ie][im]))]
                        suberr += [roccor.kSmearMCerror(int(charge[ie][im]), float(pt[ie][im]), float(eta[ie][im]), float(phi[ie][im]), int(nl[ie][im]), float(u[ie][im]))]
            res += [subres]
            err += [suberr]
    res = JaggedArray.fromiter(res)
    err = JaggedArray.fromiter(err)
    outres = res.flatten()
    outerr = err.flatten()
    np.save(fname.replace('.root','_rochester.npy'), outres)
    np.save(fname.replace('.root','_rochester_err.npy'), outerr)
    if not isData:
        outrand = np.concatenate([ui.flatten() for ui in fullu])
        np.save(fname.replace('.root','_rochester_rand.npy'), outrand)



"""A framework for analysis scale-out


"""
from .processor import ProcessorABC
from .dataframe import (
    LazyDataFrame,
    PreloadedDataFrame,
)
from .helpers import Weights, PackedSelection
from .executor import (
    iterative_executor,
    futures_executor,
    dask_executor,
    parsl_executor,
    work_queue_executor,
    run_uproot_job,
    run_parsl_job,
    run_spark_job
)
from .accumulator import (
    AccumulatorABC,
    value_accumulator,
    list_accumulator,
    set_accumulator,
    dict_accumulator,
    defaultdict_accumulator,
    column_accumulator,
)

__all__ = [
    'ProcessorABC',
    'LazyDataFrame',
    'PreloadedDataFrame',
    'Weights',
    'PackedSelection',
    'iterative_executor',
    'futures_executor',
    'dask_executor',
    'parsl_executor',
    'work_queue_executor',
    'run_uproot_job',
    'run_parsl_job',
    'run_spark_job',
    'AccumulatorABC',
    'value_accumulator',
    'list_accumulator',
    'set_accumulator',
    'dict_accumulator',
    'defaultdict_accumulator',
    'column_accumulator',
]



import numpy
import numba
import awkward1


@numba.njit
def _distinctParent_kernel(allpart_parent, allpart_pdg):
    out = numpy.empty(len(allpart_pdg), dtype=numpy.int64)
    for i in range(len(allpart_pdg)):
        parent = allpart_parent[i]
        if parent < 0:
            out[i] = -1
            continue
        thispdg = allpart_pdg[i]
        while parent >= 0 and allpart_pdg[parent] == thispdg:
            if parent >= len(allpart_pdg):
                raise RuntimeError("parent index beyond length of array!")
            parent = allpart_parent[parent]
        out[i] = parent
    return out


def distinctParent(parents, pdg):
    def gen():
        return _distinctParent_kernel(
            awkward1.Array(parents.array), awkward1.Array(pdg.array)
        )

    form = awkward1.forms.Form.fromjson('"int64"')
    return awkward1.layout.ArrayGenerator(
        gen, (), {}, form=form, length=parents.generator.length,
    )


@numba.njit
def _children_kernel(offsets_in, parentidx):
    offsets1_out = numpy.empty(len(parentidx) + 1, dtype=numpy.int64)
    content1_out = numpy.empty(len(parentidx), dtype=numpy.int64)
    offsets1_out[0] = 0

    offset0 = 0
    offset1 = 0
    for record_index in range(len(offsets_in) - 1):
        start_src, stop_src = offsets_in[record_index], offsets_in[record_index + 1]

        for index in range(start_src, stop_src):
            for possible_child in range(index, stop_src):
                if parentidx[possible_child] == index:
                    content1_out[offset1] = possible_child
                    offset1 = offset1 + 1
                    if offset1 >= len(content1_out):
                        raise RuntimeError("offset1 went out of bounds!")
            offsets1_out[offset0 + 1] = offset1
            offset0 = offset0 + 1
            if offset0 >= len(offsets1_out):
                raise RuntimeError("offset0 went out of bounds!")

    return offsets1_out, content1_out[:offset1]


def children(offsets, parents):
    def gen():
        coffsets, ccontent = _children_kernel(
            awkward1.Array(offsets.array), awkward1.Array(parents.array)
        )
        return awkward1.layout.ListOffsetArray64(
            awkward1.layout.Index64(coffsets), awkward1.layout.NumpyArray(ccontent),
        )

    form = awkward1.forms.Form.fromjson(
        '{"class": "ListOffsetArray64", "offsets": "i64", "content": "int64"}'
    )
    return awkward1.layout.ArrayGenerator(
        gen, (), {}, form=form, length=parents.generator.length,
    )



import numpy
import awkward1
from coffea.nanoevents.methods.mixin import mixin_class, mixin_method


@mixin_class
class TwoVector:
    """A cartesian 2-dimensional vector

    A heavy emphasis towards a momentum vector interpretation is assumed.
    Properties this class requires: x, y
    """

    @property
    def r(self):
        return numpy.sqrt(self.r2)

    @property
    def phi(self):
        return numpy.arctan2(self.y, self.x)

    @property
    def px(self):
        return self.x

    @property
    def py(self):
        return self.y

    @property
    def r2(self):
        return self.x ** 2 + self.y ** 2

    @property
    def pt2(self):
        return self.r2

    @property
    def pt(self):
        return self.r

    @mixin_method(numpy.absolute)
    def abs(self):
        return self.r

    @mixin_method(numpy.add, {"TwoVector"})
    def add(self, other):
        return awkward1.zip(
            {"x": self.x + other.x, "y": self.y + other.y}, with_name="TwoVector",
        )

    def sum(self, axis=-1):
        return awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
            },
            with_name="TwoVector",
        )

    @mixin_method(numpy.prod, {"float"})
    def prod(self, other):
        return awkward1.zip(
            {"x": self.x * other, "y": self.y * other}, with_name="TwoVector",
        )

    def delta_phi(self, other):
        """Compute difference in angle

        Returns a value within [-pi, pi)
        """
        return (self.phi - other.phi + numpy.pi) % (2 * numpy.pi) - numpy.pi


@mixin_class
class PolarTwoVector(TwoVector):
    """A polar coordinate 2-d vector

    This class overloads the properties r, phi with getitem accessors
    and provides properties x, y
    Some additional properties are overridden for performance
    """

    @property
    def x(self):
        return self.r * numpy.cos(self.phi)

    @property
    def y(self):
        return self.r * numpy.sin(self.phi)

    @property
    def r(self):
        return self["r"]

    @property
    def phi(self):
        return self["phi"]

    @property
    def r2(self):
        return self.r ** 2


@mixin_class
class ThreeVector(TwoVector):
    """A cartesian 3-dimensional vector

    A heavy emphasis towards a momentum vector interpretation is assumed.
    Properties this class requires: x, y, z
    """

    @property
    def pz(self):
        return self.z

    @property
    def rho2(self):
        return self.r2 + self.z ** 2

    @property
    def rho(self):
        return numpy.sqrt(self.rho2)

    @property
    def theta(self):
        return numpy.arctan2(self.r, self.z)

    @property
    def p2(self):
        return self.rho2

    @property
    def p(self):
        return self.rho

    @mixin_method(numpy.absolute)
    def abs(self):
        return self.p

    @mixin_method(numpy.add, {"ThreeVector"})
    def add(self, other):
        return awkward1.zip(
            {"x": self.x + other.x, "y": self.y + other.y, "z": self.z + other.z},
            with_name="ThreeVector",
        )

    def sum(self, axis=-1):
        return awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
                "z": awkward1.sum(self.z, axis=axis),
            },
            with_name="ThreeVector",
        )

    @mixin_method(numpy.prod, {"float"})
    def prod(self, other):
        return awkward1.zip(
            {"x": self.x * other, "y": self.y * other, "z": self.z * other},
            with_name="ThreeVector",
        )


@mixin_class
class SphericalThreeVector(ThreeVector, PolarTwoVector):
    """A spherical coordinate 3-d vector

    This class overloads the properties rho, theta with getitem accessors
    and provides properties r, z
    Some additional properties are overridden for performance
    """

    @property
    def r(self):
        return self.rho * numpy.sin(self.theta)

    @property
    def z(self):
        return self.rho * numpy.cos(self.theta)

    @property
    def rho(self):
        return self["rho"]

    @property
    def theta(self):
        return self["theta"]

    @property
    def p(self):
        return self.rho

    @property
    def p2(self):
        return self.rho ** 2


@mixin_class
class LorentzVector(ThreeVector):
    """A cartesian Lorentz vector

    A heavy emphasis towards a momentum vector interpretation is assumed.
    + - - - metric
    Properties this class requires: x, y, z, t
    """

    @property
    def energy(self):
        return self.t

    @property
    def eta(self):
        return numpy.arcsinh(self.z / self.r)

    @property
    def mass2(self):
        return self.t ** 2 - self.p2

    @property
    def mass(self):
        return numpy.sqrt(self.mass2)

    @mixin_method(numpy.absolute)
    def abs(self):
        return self.mass

    @mixin_method(numpy.add, {"LorentzVector"})
    def add(self, other):
        return awkward1.zip(
            {
                "x": self.x + other.x,
                "y": self.y + other.y,
                "z": self.z + other.z,
                "t": self.t + other.t,
            },
            with_name="LorentzVector",
        )

    def sum(self, axis=-1):
        return awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
                "z": awkward1.sum(self.z, axis=axis),
                "t": awkward1.sum(self.t, axis=axis),
            },
            with_name="LorentzVector",
        )

    @mixin_method(numpy.prod, {"float"})
    def prod(self, other):
        return awkward1.zip(
            {
                "x": self.x * other,
                "y": self.y * other,
                "z": self.z * other,
                "t": self.t * other,
            },
            with_name="LorentzVector",
        )

    def delta_r2(self, other):
        return (self.eta - other.eta) ** 2 + self.delta_phi(other) ** 2

    def delta_r(self, other):
        return numpy.sqrt(self.delta_r2(other))

    def nearest(self, other, metric=lambda a, b: a.delta_r(b), return_metric=False):
        """Return nearest object to this one

        Only works for first axis (i.e. top-level ListArrays)
        """
        a, b = awkward1.unzip(awkward1.cartesian([self, other], nested=True))
        mval = metric(a, b)
        mmin = awkward1.argmin(mval, axis=-1)
        if return_metric:
            return b[mmin], mval[mmin]
        return b[mmin]


@mixin_class
class PtEtaPhiMLorentzVector(LorentzVector, SphericalThreeVector):
    """A Lorentz vector using pseudorapidity and mass

    This class overloads the pt, eta, phi, mass  properties with getitem accessors
    and provides properties rho, theta, r, z, t
    Some additional properties are overridden for performance
    """

    @property
    def pt(self):
        return self["pt"]

    @property
    def eta(self):
        return self["eta"]

    @property
    def phi(self):
        return self["phi"]

    @property
    def mass(self):
        return self["mass"]

    @property
    def rho(self):
        return self.pt * numpy.cosh(self.eta)

    @property
    def theta(self):
        raise NotImplementedError

    @property
    def r(self):
        return self.pt

    @property
    def z(self):
        return self.pt * numpy.sinh(self.eta)

    @property
    def t(self):
        return numpy.hypot(self.rho, self.mass)

    @property
    def rho2(self):
        return self.rho ** 2

    @property
    def mass2(self):
        return self.mass ** 2

    @mixin_method(numpy.prod, {"float"})
    def prod(self, other):
        return awkward1.zip(
            {
                "pt": self.pt * other,
                "eta": self.eta,
                "phi": self.phi,
                "mass": self.mass * other,
            },
            with_name="PtEtaPhiMLorentzVector",
        )



from weakref import WeakValueDictionary
import numpy
import awkward1
import uproot
from coffea.nanoevents.indirection import distinctParent, children


def _with_length(array: awkward1.layout.VirtualArray, length: int):
    return awkward1.layout.VirtualArray(
        array.generator.with_length(length),
        array.cache,
        array.cache_key,
        array.identities,
        array.parameters,
    )


class NanoEventsFactory:
    """A factory class to build NanoEvents objects

    This factory must last for the lifetime of any events or sub-events objects
    for indirections (such as cross-references) to remain valid. Data members and
    mixins will work ok if the factory goes out of scope, however.

    Parameters
    ----------
        file : str or uproot.rootio.ROOTDirectory
            The filename or already opened file using e.g. ``uproot.open()``
        treename : str, optional
            Name of the tree to read in the file, defaults to ``Events``
        entrystart : int, optional
            Start at this entry offset in the tree (default 0)
        entrystop : int, optional
            Stop at this entry offset in the tree (default end of tree)
        cache : dict, optional
            A dict-like interface to a cache object, in which any materialized virtual arrays will be kept
        mixin_map : dict, optional
            A mapping from collection name to mixin class name that implements custom
            additional mixins beyond the defaults provided.
        metadata : dict, optional
            Arbitrary metadata to embed in this NanoEvents, accessible via ``events.metadata``
    """

    default_mixins = {
        "CaloMET": "MissingET",
        "ChsMET": "MissingET",
        "GenMET": "MissingET",
        "MET": "MissingET",
        "METFixEE2017": "MissingET",
        "PuppiMET": "MissingET",
        "RawMET": "MissingET",
        "TkMET": "MissingET",
        # pseudo-lorentz: pt, eta, phi, mass=0
        "IsoTrack": "PtEtaPhiMCollection",
        "SoftActivityJet": "PtEtaPhiMCollection",
        "TrigObj": "PtEtaPhiMCollection",
        # True lorentz: pt, eta, phi, mass
        "FatJet": "FatJet",
        "GenDressedLepton": "PtEtaPhiMCollection",
        "GenJet": "PtEtaPhiMCollection",
        "GenJetAK8": "FatJet",
        "Jet": "Jet",
        "LHEPart": "PtEtaPhiMCollection",
        "SV": "PtEtaPhiMCollection",
        "SubGenJetAK8": "PtEtaPhiMCollection",
        "SubJet": "PtEtaPhiMCollection",
        # Candidate: lorentz + charge
        "Electron": "Electron",
        "Muon": "Muon",
        "Photon": "Photon",
        "Tau": "Tau",
        "GenVisTau": "GenVisTau",
        # special
        "GenPart": "GenParticle",
    }
    """Default configuration for mixin types, based on the collection name."""
    nested_items = {
        "FatJet_subJetIdxG": ["FatJet_subJetIdx1G", "FatJet_subJetIdx2G"],
        "Jet_muonIdxG": ["Jet_muonIdx1G", "Jet_muonIdx2G"],
        "Jet_electronIdxG": ["Jet_electronIdx1G", "Jet_electronIdx2G"],
    }
    """Default nested collections, where nesting is accomplished by a fixed-length set of indexers"""
    special_items = {
        "GenPart_distinctParentIdxG": (
            distinctParent,
            ("GenPart_genPartIdxMotherG", "GenPart_pdgId"),
        ),
        "GenPart_childrenIdxG": (children, ("oGenPart", "GenPart_genPartIdxMotherG")),
        "GenPart_distinctChildrenIdxG": (
            children,
            ("oGenPart", "GenPart_distinctParentIdxG"),
        ),
    }
    """Default special arrays, where the callable and input arrays are specified in the value"""
    _active = WeakValueDictionary()

    def __init__(
        self,
        file,
        treename="Events",
        entrystart=None,
        entrystop=None,
        cache=None,
        mixin_map=None,
        metadata=None,
    ):
        if not isinstance(file, uproot.rootio.ROOTDirectory):
            file = uproot.open(file)
        self._tree = file[treename]
        self._entrystart, self._entrystop = uproot.tree._normalize_entrystartstop(
            self._tree.numentries, entrystart, entrystop
        )
        self._keyprefix = "/".join(
            [
                file._context.uuid.hex(),
                treename,
                str(self._entrystart),
                str(self._entrystop),
            ]
        )
        NanoEventsFactory._active[self._keyprefix] = self

        if cache is None:
            cache = awkward1.layout.ArrayCache({})
        else:
            cache = awkward1.layout.ArrayCache(cache)
        self._cache = cache

        self._mixin_map = {}
        self._mixin_map.update(self.default_mixins)
        if mixin_map is not None:
            self._mixin_map.update(mixin_map)

        self._metadata = metadata  # TODO: JSON only?
        self._branches_read = set()
        self._collections = None

    @classmethod
    def _instance(cls, key):
        try:
            return cls._active[key]
        except KeyError:
            raise RuntimeError(
                "NanoEventsFactory instance was lost, cross-references are now invalid"
            )

    @classmethod
    def get_events(cls, key):
        """The events corresponding to the factory keyed by ``key``

        This is used mainly for indirections
        """
        return cls._instance(key).events()

    @classmethod
    def get_cache(cls, key):
        """The cache corresponding to the factory keyed by ``key``"""
        return cls._instance(key)._cache

    def __len__(self):
        return self._entrystop - self._entrystart

    def reader(self, branch_name, parameters):
        """Read a branch from the ROOT file as an awkward1 array"""
        self._branches_read.add(branch_name)
        return awkward1.layout.NumpyArray(
            self._tree[branch_name].array(
                entrystart=self._entrystart, entrystop=self._entrystop, flatten=True
            ),
            parameters=parameters,
        )

    def _array(self, branch_name):
        interpretation = uproot.interpret(self._tree[branch_name])
        if isinstance(interpretation, uproot.asjagged):
            dtype = interpretation.content.type
            length = None
        else:
            dtype = interpretation.type
            length = len(self)
        parameters = {
            "__doc__": self._tree[branch_name].title.decode("ascii"),
            "__record__": "NanoColumn",
        }
        # use hint to resolve platform-dependent format
        formhint = awkward1.forms.Form.fromjson('"%s"' % dtype)
        form = awkward1.forms.NumpyForm(
            [], formhint.itemsize, formhint.format, parameters=parameters
        )
        generator = awkward1.layout.ArrayGenerator(
            self.reader, (branch_name, parameters), {}, form=form, length=length,
        )
        source = "file"
        return awkward1.layout.VirtualArray(
            generator,
            self._cache,
            "/".join([self._keyprefix, source, branch_name]),
            parameters=parameters,
        )

    def _getoffsets(self, counts_name, parameters):
        counts = self.reader(counts_name, parameters)
        offsets = numpy.empty(len(counts) + 1, numpy.uint32)
        offsets[0] = 0
        numpy.cumsum(counts, out=offsets[1:])
        return awkward1.layout.NumpyArray(offsets, parameters=parameters,)

    def _counts2offsets(self, virtual_counts, name):
        generator = virtual_counts.generator
        generator = generator.with_callable(self._getoffsets).with_length(
            generator.length + 1
        )
        return awkward1.layout.VirtualArray(
            generator,
            self._cache,
            "/".join([self._keyprefix, "counts2offsets", name]),
            virtual_counts.identities,
            virtual_counts.parameters,
        )

    def _local2global(self, index, source_offsets, target_offsets, name):
        def globalindex():
            gidx = awkward1.Array(
                awkward1.layout.ListOffsetArray32(
                    awkward1.layout.Index32(source_offsets), index.generator(),
                )
            )
            gidx = gidx.mask[gidx >= 0] + target_offsets[:-1]
            return awkward1.fill_none(awkward1.flatten(gidx), -1)

        generator = awkward1.layout.ArrayGenerator(
            globalindex,
            (),
            {},
            form=awkward1.forms.Form.fromjson('"int64"'),
            length=index.generator.length,
        )
        return awkward1.layout.VirtualArray(
            generator,
            index.cache,
            "/".join([self._keyprefix, "local2global", name]),
            index.identities,
            index.parameters,
        )

    def _nestedindex(self, indexers, name):
        def nestedindex():
            # idx = awkward1.concatenate([idx[:, None] for idx in indexers], axis=1)
            n = len(indexers)
            out = numpy.empty(n * len(indexers[0]), dtype="int64")
            for i, idx in enumerate(indexers):
                out[i::n] = idx
            offsets = numpy.arange(0, len(out) + 1, n, dtype=numpy.int64)
            return awkward1.layout.ListOffsetArray64(
                awkward1.layout.Index64(offsets), awkward1.layout.NumpyArray(out),
            )

        form = awkward1.forms.Form.fromjson(
            '{"class": "ListOffsetArray64", "offsets": "i64", "content": "int64"}'
        )
        generator = awkward1.layout.ArrayGenerator(
            nestedindex, (), {}, form=form, length=indexers[0].generator.length,
        )
        return awkward1.layout.VirtualArray(
            generator,
            self._cache,
            cache_key="/".join([self._keyprefix, "nestedindex", name]),
        )

    def _listarray(self, offsets, content, params):
        offsets = awkward1.layout.Index32(offsets)
        length = offsets[-1]
        if isinstance(content, dict):
            content = awkward1.layout.RecordArray(
                {k: _with_length(v, length) for k, v in content.items()},
                parameters=params,
            )
            return awkward1.layout.ListOffsetArray32(offsets, content)
        return awkward1.layout.ListOffsetArray32(offsets, content, parameters=params)

    def _build_collections(self):
        if self._collections is not None:
            return self._collections

        arrays = {branch_name.decode("ascii") for branch_name in self._tree.keys()}

        # parse into high-level records (collections, list collections, and singletons)
        collections = set(k.split("_")[0] for k in arrays)
        collections -= set(
            k for k in collections if k.startswith("n") and k[1:] in collections
        )

        arrays = {k: self._array(k) for k in arrays}

        # Create offsets virtual arrays
        for name in collections:
            if "n" + name in arrays:
                arrays["o" + name] = self._counts2offsets(
                    arrays["n" + name], "o" + name
                )

        # Create global index virtual arrays for indirection
        for name in collections:
            indexers = filter(lambda k: k.startswith(name) and "Idx" in k, arrays)
            for k in list(indexers):
                target = k[len(name) + 1 : k.find("Idx")]
                target = target[0].upper() + target[1:]
                if target not in collections:
                    raise RuntimeError(
                        "Parsing indexer %s, expected to find collection %s but did not"
                        % (k, target)
                    )
                arrays[k + "G"] = self._local2global(
                    arrays[k], arrays["o" + name], arrays["o" + target], k + "G"
                )

        # Create nested indexer from Idx1, Idx2, ... arrays
        for name, indexers in self.nested_items.items():
            if all(idx in arrays for idx in indexers):
                arrays[name] = self._nestedindex(
                    [arrays[idx] for idx in indexers], name
                )

        # Create any special arrays
        for name, (fcn, args) in self.special_items.items():
            if all(k in arrays for k in args):
                generator = fcn(*(arrays[k] for k in args))
                arrays[name] = awkward1.layout.VirtualArray(
                    generator,
                    self._cache,
                    cache_key="/".join([self._keyprefix, fcn.__name__, name]),
                )

        def collectionfactory(name):
            mixin = self._mixin_map.get(name, "NanoCollection")
            if "o" + name in arrays and name not in arrays:
                # list collection
                offsets = arrays["o" + name]
                content = {
                    k[len(name) + 1 :]: arrays[k]
                    for k in arrays
                    if k.startswith(name + "_")
                }
                recordparams = {
                    "__doc__": offsets.parameters["__doc__"],
                    "__record__": mixin,
                    "events_key": self._keyprefix,
                    "collection_name": name,
                }
                form = awkward1.forms.ListOffsetForm(
                    "i32",
                    awkward1.forms.RecordForm(
                        {k: v.form for k, v in content.items()}, parameters=recordparams
                    ),
                )
                generator = awkward1.layout.ArrayGenerator(
                    self._listarray,
                    (offsets, content, recordparams),
                    {},
                    form=form,
                    length=len(self),
                )
                source = "runtime"
                return awkward1.layout.VirtualArray(
                    generator,
                    self._cache,
                    cache_key="/".join([self._keyprefix, source, name]),
                    parameters=recordparams,
                )
            elif "o" + name in arrays:
                # list singleton
                offsets = arrays["o" + name]
                content = arrays[name]
                params = {
                    # This makes more sense as offsets doc but it seems that is empty
                    "__doc__": content.parameters["__doc__"],
                    "__array__": mixin,
                    "events_key": self._keyprefix,
                    "collection_name": name,
                }
                form = awkward1.forms.ListOffsetForm(
                    "i32", content.form, parameters=params
                )
                generator = awkward1.layout.ArrayGenerator(
                    self._listarray,
                    (offsets, content, params),
                    {},
                    form=form,
                    length=len(self),
                )
                source = "runtime"
                return awkward1.layout.VirtualArray(
                    generator,
                    self._cache,
                    cache_key="/".join([self._keyprefix, source, name]),
                    parameters=params,
                )
            elif name in arrays:
                # singleton
                return arrays[name]
            else:
                # simple collection
                content = {
                    k[len(name) + 1 :]: arrays[k]
                    for k in arrays
                    if k.startswith(name + "_")
                }
                return awkward1.layout.RecordArray(
                    content,
                    parameters={
                        "__record__": mixin,
                        "events_key": self._keyprefix,
                        "collection_name": name,
                    },
                )

        self._collections = {name: collectionfactory(name) for name in collections}

    def events(self):
        """Build events

        The NanoEvents object is built from all branches found in the supplied file, based on
        the naming pattern of the branches. The following additional arrays are constructed:

        - Any branches named ``n{name}`` are assumed to be counts branches and converted to offsets ``o{name}``
        - Any local index branches with names matching ``{source}_{target}Idx*`` are converted to global indexes for the event chunk
        - Any `NanoEventsFactory.nested_items` are constructed, if the necessary branches are available
        - Any `NanoEventsFactory.special_items` are constructed, if the necessary branches are available

        From those arrays, NanoAOD collections are formed as collections of branches grouped by name, where:

        - one branch exists named ``name`` and no branches start with ``name_``, interpreted as a single flat array;
        - one branch exists named ``name``, one named ``n{name}``, and no branches start with ``name_``, interpreted as a single jagged array;
        - no branch exists named ``{name}`` and many branches start with ``name_*``, interpreted as a flat table; or
        - one branch exists named ``n{name}`` and many branches start with ``name_*``, interpreted as a jagged table.

        All collections are then zipped into one `NanoEvents` record and returned.
        """
        self._build_collections()
        events = awkward1.layout.RecordArray(
            self._collections,
            parameters={
                "__record__": "NanoEvents",
                "__doc__": self._tree.title.decode("ascii"),
                "events_key": self._keyprefix,
                "metadata": self._metadata,
            },
        )
        return awkward1.Array(events)



# -*- coding: utf-8 -*-
from __future__ import division
from ..util import numpy as np
import scipy.stats
import copy
import warnings
import numbers
from .hist_tools import SparseAxis, DenseAxis, overflow_behavior, Interval, StringBin

# Plotting is always terrible
# Let's try our best to follow matplotlib idioms
# https://matplotlib.org/tutorials/introductory/usage.html#coding-styles

_coverage1sd = scipy.stats.norm.cdf(1) - scipy.stats.norm.cdf(-1)


def poisson_interval(sumw, sumw2, coverage=_coverage1sd):
    """Frequentist coverage interval for Poisson-distributed observations

    Parameters
    ----------
        sumw : numpy.ndarray
            Sum of weights vector
        sumw2 : numpy.ndarray
            Sum weights squared vector
        coverage : float, optional
            Central coverage interval, defaults to 68%

    Calculates the so-called 'Garwood' interval,
    c.f. https://www.ine.pt/revstat/pdf/rs120203.pdf or
    http://ms.mcmaster.ca/peter/s743/poissonalpha.html
    For weighted data, this approximates the observed count by ``sumw**2/sumw2``, which
    effectively scales the unweighted poisson interval by the average weight.
    This may not be the optimal solution: see https://arxiv.org/pdf/1309.1287.pdf for a proper treatment.
    When a bin is zero, the scale of the nearest nonzero bin is substituted to scale the nominal upper bound.
    If all bins zero, a warning is generated and interval is set to ``sumw``.
    """
    scale = np.empty_like(sumw)
    scale[sumw != 0] = sumw2[sumw != 0] / sumw[sumw != 0]
    if np.sum(sumw == 0) > 0:
        missing = np.where(sumw == 0)
        available = np.nonzero(sumw)
        if len(available[0]) == 0:
            warnings.warn("All sumw are zero!  Cannot compute meaningful error bars", RuntimeWarning)
            return np.vstack([sumw, sumw])
        nearest = sum([np.subtract.outer(d, d0)**2 for d, d0 in zip(available, missing)]).argmin(axis=0)
        argnearest = tuple(dim[nearest] for dim in available)
        scale[missing] = scale[argnearest]
    counts = sumw / scale
    lo = scale * scipy.stats.chi2.ppf((1 - coverage) / 2, 2 * counts) / 2.
    hi = scale * scipy.stats.chi2.ppf((1 + coverage) / 2, 2 * (counts + 1)) / 2.
    interval = np.array([lo, hi])
    interval[interval == np.nan] = 0.  # chi2.ppf produces nan for counts=0
    return interval


def clopper_pearson_interval(num, denom, coverage=_coverage1sd):
    """Compute Clopper-Pearson coverage interval for a binomial distribution

    Parameters
    ----------
        num : numpy.ndarray
            Numerator, or number of successes, vectorized
        denom : numpy.ndarray
            Denominator or number of trials, vectorized
        coverage : float, optional
            Central coverage interval, defaults to 68%

    c.f. http://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval
    """
    if np.any(num > denom):
        raise ValueError("Found numerator larger than denominator while calculating binomial uncertainty")
    lo = scipy.stats.beta.ppf((1 - coverage) / 2, num, denom - num + 1)
    hi = scipy.stats.beta.ppf((1 + coverage) / 2, num + 1, denom - num)
    interval = np.array([lo, hi])
    interval[:, num == 0.] = 0.
    interval[1, num == denom] = 1.
    return interval


def normal_interval(pw, tw, pw2, tw2, coverage=_coverage1sd):
    """Compute errors based on the expansion of pass/(pass + fail), possibly weighted

        Parameters
        ----------
        pw : numpy.ndarray
            Numerator, or number of (weighted) successes, vectorized
        tw : numpy.ndarray
            Denominator or number of (weighted) trials, vectorized
        pw2 : numpy.ndarray
            Numerator sum of weights squared, vectorized
        tw2 : numpy.ndarray
            Denominator sum of weights squared, vectorized
        coverage : float, optional
            Central coverage interval, defaults to 68%

        c.f. https://root.cern.ch/doc/master/TEfficiency_8cxx_source.html#l02515
    """

    eff = pw / tw

    variance = (pw2 * (1 - 2 * eff) + tw2 * eff**2) / (tw**2)
    sigma = np.sqrt(variance)

    prob = 0.5 * (1 - coverage)
    delta = np.zeros_like(sigma)
    delta[sigma != 0] = scipy.stats.norm.ppf(prob, scale=sigma[sigma != 0])

    lo = eff - np.minimum(eff + delta, np.ones_like(eff))
    hi = np.maximum(eff - delta, np.zeros_like(eff)) - eff

    return np.array([lo, hi])


def plot1d(hist, ax=None, clear=True, overlay=None, stack=False, overflow='none', line_opts=None,
           fill_opts=None, error_opts=None, legend_opts={}, overlay_overflow='none',
           density=False, binwnorm=None, order=None):
    """Create a 1D plot from a 1D or 2D `Hist` object

    Parameters
    ----------
        hist : Hist
            Histogram with maximum of two dimensions
        ax : matplotlib.axes.Axes, optional
            Axes object (if None, one is created)
        clear : bool, optional
            Whether to clear Axes before drawing (if passed); if False, this function will skip drawing the legend
        overlay : str, optional
            In the case that ``hist`` is 2D, specify the axis of hist to overlay (remaining axis will be x axis)
        stack : bool, optional
            Whether to stack or overlay non-axis dimension (if it exists)
        order : list, optional
            How to order when stacking. Take a list of identifiers.
        overflow : str, optional
            If overflow behavior is not 'none', extra bins will be drawn on either end of the nominal
            axis range, to represent the contents of the overflow bins.  See `Hist.sum` documentation
            for a description of the options.
        line_opts : dict, optional
            A dictionary of options to pass to the matplotlib
            `ax.step <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.step.html>`_ call
            internal to this function.  Leave blank for defaults.
        fill_opts : dict, optional
            A dictionary of options to pass to the matplotlib
            `ax.fill_between <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.fill_between.html>`_ call
            internal to this function.  Leave blank for defaults.
        error_opts : dict, optional
            A dictionary of options to pass to the matplotlib
            `ax.errorbar <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.errorbar.html>`_ call
            internal to this function.  Leave blank for defaults.  Some special options are interpreted by
            this function and not passed to matplotlib: 'emarker' (default: '') specifies the marker type
            to place at cap of the errorbar.
        legend_opts : dict, optional
            A dictionary of options  to pass to the matplotlib
            `ax.legend <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.legend.html>`_ call
            internal to this fuction.  Leave blank for defaults.
        overlay_overflow : str, optional
            If overflow behavior is not 'none', extra bins in the overlay axis will be overlayed or stacked,
            to represent the contents of the overflow bins.  See `Hist.sum` documentation for a description of the options.
        density : bool, optional
            If true, convert sum weights to probability density (i.e. integrates to 1 over domain of axis)
            (Note: this option conflicts with ``binwnorm``)
        binwnorm : float, optional
            If true, convert sum weights to bin-width-normalized, with unit equal to supplied value (usually you want to specify 1.)


    Returns
    -------
        ax : matplotlib.axes.Axes
            A matplotlib `Axes <https://matplotlib.org/3.1.1/api/axes_api.html>`_ object
    """
    import mplhep as hep
    import matplotlib.pyplot as plt
    if ax is None:
        ax = plt.gca()
    else:
        if not isinstance(ax, plt.Axes):
            raise ValueError("ax must be a matplotlib Axes object")
        if clear:
            ax.clear()
    if hist.dim() > 2:
        raise ValueError("plot1d() can only support up to two dimensions (one for axis, one to stack or overlay)")
    if overlay is None and hist.sparse_dim() == 1 and hist.dense_dim() == 1:
        overlay = hist.sparse_axes()[0].name
    elif overlay is None and hist.dim() > 1:
        raise ValueError("plot1d() can only support one dimension without an overlay axis chosen")
    if density and binwnorm is not None:
        raise ValueError("Cannot use density and binwnorm at the same time!")
    if binwnorm is not None:
        if not isinstance(binwnorm, numbers.Number):
            raise ValueError("Bin width normalization not a number, but a %r" % binwnorm.__class__)
    if line_opts is None and fill_opts is None and error_opts is None:
        if stack:
            fill_opts = {}
        else:
            line_opts = {}
            error_opts = {}

    axis = hist.axes()[0]
    if overlay is not None:
        overlay = hist.axis(overlay)
        if axis == overlay:
            axis = hist.axes()[1]
    if isinstance(axis, SparseAxis):
        raise NotImplementedError("Plot a sparse axis (e.g. bar chart)")
    elif isinstance(axis, DenseAxis):
        ax.set_xlabel(axis.label)
        ax.set_ylabel(hist.label)
        edges = axis.edges(overflow=overflow)
        if order is None:
            identifiers = hist.identifiers(overlay, overflow=overlay_overflow) if overlay is not None else [None]
        else:
            identifiers = order
        plot_info = {
            'identifier': identifiers,
            'label': list(map(str, identifiers)),
            'sumw': [],
            'sumw2': []
        }
        for i, identifier in enumerate(identifiers):
            if identifier is None:
                sumw, sumw2 = hist.values(sumw2=True, overflow=overflow)[()]
            elif isinstance(overlay, SparseAxis):
                sumw, sumw2 = hist.integrate(overlay, identifier).values(sumw2=True, overflow=overflow)[()]
            else:
                sumw, sumw2 = hist.values(sumw2=True, overflow='allnan')[()]
                the_slice = (i if overflow_behavior(overlay_overflow).start is None else i + 1, overflow_behavior(overflow))
                if hist._idense(overlay) == 1:
                    the_slice = (the_slice[1], the_slice[0])
                sumw = sumw[the_slice]
                sumw2 = sumw2[the_slice]
            plot_info['sumw'].append(sumw)
            plot_info['sumw2'].append(sumw2)

        def w2err(sumw, sumw2):
            err = []
            for a, b in zip(sumw, sumw2):
                err.append(np.abs(poisson_interval(a, b) - a))
            return err

        kwargs = None
        if line_opts is not None and error_opts is None:
            _error = None
        else:
            _error = w2err(plot_info['sumw'], plot_info['sumw2'])
        if fill_opts is not None:
            histtype = 'fill'
            kwargs = fill_opts
        elif error_opts is not None and line_opts is None:
            histtype = 'errorbar'
            kwargs = error_opts
        else:
            histtype = 'step'
            kwargs = line_opts
        if kwargs is None:
            kwargs = {}

        hep.histplot(plot_info['sumw'], edges, label=plot_info['label'],
                     yerr=_error, histtype=histtype, ax=ax,
                     density=density, binwnorm=binwnorm, stack=stack,
                     **kwargs)

        if stack and error_opts is not None:
            stack_sumw = np.sum(plot_info['sumw'], axis=0)
            stack_sumw2 = np.sum(plot_info['sumw2'], axis=0)
            err = poisson_interval(stack_sumw, stack_sumw2)
            if binwnorm is not None:
                err *= binwnorm / np.diff(edges)[None, :]
            opts = {'step': 'post', 'label': 'Sum unc.', 'hatch': '///',
                    'facecolor': 'none', 'edgecolor': (0, 0, 0, .5), 'linewidth': 0}
            opts.update(error_opts)
            ax.fill_between(x=edges, y1=np.r_[err[0, :], err[0, -1]],
                            y2=np.r_[err[1, :], err[1, -1]], **opts)

        if legend_opts is not None:
            _label = overlay.label if overlay is not None else ""
            ax.legend(title=_label, **legend_opts)
        else:
            ax.legend(title=_label)
        ax.autoscale(axis='x', tight=True)
        ax.set_ylim(0, None)

    return ax


def plotratio(num, denom, ax=None, clear=True, overflow='none', error_opts=None, denom_fill_opts=None, guide_opts=None, unc='clopper-pearson', label=None):
    """Create a ratio plot, dividing two compatible histograms

    Parameters
    ----------
        num : Hist
            Numerator, a single-axis histogram
        denom : Hist
            Denominator, a single-axis histogram
        ax : matplotlib.axes.Axes, optional
            Axes object (if None, one is created)
        clear : bool, optional
            Whether to clear Axes before drawing (if passed); if False, this function will skip drawing the legend
        overflow : str, optional
            If overflow behavior is not 'none', extra bins will be drawn on either end of the nominal
            axis range, to represent the contents of the overflow bins.  See `Hist.sum` documentation
            for a description of the options.
        error_opts : dict, optional
            A dictionary of options to pass to the matplotlib
            `ax.errorbar <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.errorbar.html>`_ call
            internal to this function.  Leave blank for defaults.  Some special options are interpreted by
            this function and not passed to matplotlib: 'emarker' (default: '') specifies the marker type
            to place at cap of the errorbar.
        denom_fill_opts : dict, optional
            A dictionary of options to pass to the matplotlib
            `ax.fill_between <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.fill_between.html>`_ call
            internal to this function, filling the denominator uncertainty band.  Leave blank for defaults.
        guide_opts : dict, optional
            A dictionary of options to pass to the matplotlib
            `ax.axhline <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.axhline.html>`_ call
            internal to this function, to plot a horizontal guide line at ratio of 1.  Leave blank for defaults.
        unc : str, optional
            Uncertainty calculation option: 'clopper-pearson' interval for efficiencies; 'poisson-ratio' interval
            for ratio of poisson distributions; 'num' poisson interval of numerator scaled by denominator value
            (common for data/mc, for better or worse).
        label : str, optional
            Associate a label to this entry (note: y axis label set by ``num.label``)

    Returns
    -------
        ax : matplotlib.axes.Axes
            A matplotlib `Axes <https://matplotlib.org/3.1.1/api/axes_api.html>`_ object
    """
    import mplhep as hep
    import matplotlib.pyplot as plt
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    else:
        if not isinstance(ax, plt.Axes):
            raise ValueError("ax must be a matplotlib Axes object")
        if clear:
            ax.clear()
    if not num.compatible(denom):
        raise ValueError("numerator and denominator histograms have incompatible axis definitions")
    if num.dim() > 1:
        raise ValueError("plotratio() can only support one-dimensional histograms")
    if error_opts is None and denom_fill_opts is None and guide_opts is None:
        error_opts = {}
        denom_fill_opts = {}

    axis = num.axes()[0]
    if isinstance(axis, SparseAxis):
        raise NotImplementedError("Ratio for sparse axes (labeled axis with errorbars)")
    elif isinstance(axis, DenseAxis):
        ax.set_xlabel(axis.label)
        ax.set_ylabel(num.label)
        edges = axis.edges(overflow=overflow)
        centers = axis.centers(overflow=overflow)

        sumw_num, sumw2_num = num.values(sumw2=True, overflow=overflow)[()]
        sumw_denom, sumw2_denom = denom.values(sumw2=True, overflow=overflow)[()]

        rsumw = sumw_num / sumw_denom
        if unc == 'clopper-pearson':
            rsumw_err = np.abs(clopper_pearson_interval(sumw_num, sumw_denom) - rsumw)
        elif unc == 'poisson-ratio':
            # poisson ratio n/m is equivalent to binomial n/(n+m)
            rsumw_err = np.abs(clopper_pearson_interval(sumw_num, sumw_num + sumw_denom) - rsumw)
        elif unc == 'num':
            rsumw_err = np.abs(poisson_interval(rsumw, sumw2_num / sumw_denom**2) - rsumw)
        elif unc == "normal":
            rsumw_err = np.abs(normal_interval(sumw_num, sumw_denom, sumw2_num, sumw2_denom))
        else:
            raise ValueError("Unrecognized uncertainty option: %r" % unc)

        if error_opts is not None:
            opts = {'label': label, 'linestyle': 'none'}
            opts.update(error_opts)
            emarker = opts.pop('emarker', '')
            errbar = ax.errorbar(x=centers, y=rsumw, yerr=rsumw_err, **opts)
            plt.setp(errbar[1], 'marker', emarker)
        if denom_fill_opts is not None:
            unity = np.ones_like(sumw_denom)
            denom_unc = poisson_interval(unity, sumw2_denom / sumw_denom**2)
            opts = {'step': 'post', 'facecolor': (0, 0, 0, 0.3), 'linewidth': 0}
            opts.update(denom_fill_opts)
            ax.fill_between(edges, np.r_[denom_unc[0], denom_unc[0, -1]], np.r_[denom_unc[1], denom_unc[1, -1]], **opts)
        if guide_opts is not None:
            opts = {'linestyle': '--', 'color': (0, 0, 0, 0.5), 'linewidth': 1}
            opts.update(guide_opts)
            ax.axhline(1., **opts)

    if clear:
        ax.autoscale(axis='x', tight=True)
        ax.set_ylim(0, None)

    return ax


def plot2d(hist, xaxis, ax=None, clear=True, xoverflow='none', yoverflow='none', patch_opts=None, text_opts=None, density=False, binwnorm=None):
    """Create a 2D plot from a 2D `Hist` object

    Parameters
    ----------
        hist : Hist
            Histogram with two dimensions
        xaxis : str or Axis
            Which of the two dimensions to use as an x axis
        ax : matplotlib.axes.Axes, optional
            Axes object (if None, one is created)
        clear : bool, optional
            Whether to clear Axes before drawing (if passed); if False, this function will skip drawing the legend
        xoverflow : str, optional
            If overflow behavior is not 'none', extra bins will be drawn on either end of the nominal x
            axis range, to represent the contents of the overflow bins.  See `Hist.sum` documentation
            for a description of the options.
        yoverflow : str, optional
            Similar to ``xoverflow``
        patch_opts : dict, optional
            Options passed to the matplotlib `pcolormesh <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html>`_
            call internal to this function, to plot a rectangular grid of patches colored according to the bin values.
            Leave empty for defaults.
        text_opts : dict, optional
            Options passed to the matplotlib `text <https://matplotlib.org/api/text_api.html#matplotlib.text.Text>`_
            call internal to this function, to place a text label at each bin center with the bin value.  Special
            options interpreted by this function and not passed to matplotlib: 'format': printf-style float format
            , default '%.2g'.
        density : bool, optional
            If true, convert sum weights to probability density (i.e. integrates to 1 over domain of axis)
            (Note: this option conflicts with ``binwnorm``)
        binwnorm : float, optional
            If true, convert sum weights to bin-width-normalized, with unit equal to supplied value (usually you want to specify 1.)

    Returns
    -------
        ax : matplotlib.axes.Axes
            A matplotlib `Axes <https://matplotlib.org/3.1.1/api/axes_api.html>`_ object
    """
    import mplhep as hep
    import matplotlib.pyplot as plt
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    else:
        if not isinstance(ax, plt.Axes):
            raise ValueError("ax must be a matplotlib Axes object")
        if clear:
            ax.clear()
        fig = ax.figure
    if hist.dim() != 2:
        raise ValueError("plot2d() can only support exactly two dimensions")
    if density and binwnorm is not None:
        raise ValueError("Cannot use density and binwnorm at the same time!")
    if binwnorm is not None:
        if not isinstance(binwnorm, numbers.Number):
            raise ValueError("Bin width normalization not a number, but a %r" % binwnorm.__class__)
    if patch_opts is None and text_opts is None:
        patch_opts = {}

    xaxis = hist.axis(xaxis)
    yaxis = hist.axes()[1]
    transpose = False
    if yaxis == xaxis:
        yaxis = hist.axes()[0]
        transpose = True
    if isinstance(xaxis, SparseAxis) or isinstance(yaxis, SparseAxis):
        raise NotImplementedError("Plot a sparse axis (e.g. bar chart or labeled bins)")
    else:
        xedges = xaxis.edges(overflow=xoverflow)
        yedges = yaxis.edges(overflow=yoverflow)
        sumw, sumw2 = hist.values(sumw2=True, overflow='allnan')[()]
        if transpose:
            sumw = sumw.T
            sumw2 = sumw2.T
        # no support for different overflow behavior per axis, do it ourselves
        sumw = sumw[overflow_behavior(xoverflow), overflow_behavior(yoverflow)]
        sumw2 = sumw2[overflow_behavior(xoverflow), overflow_behavior(yoverflow)]
        if (density or binwnorm is not None) and np.sum(sumw) > 0:
            overallnorm = np.sum(sumw) * binwnorm if binwnorm is not None else 1.
            areas = np.multiply.outer(np.diff(xedges), np.diff(yedges))
            binnorms = overallnorm / (areas * np.sum(sumw))
            sumw = sumw * binnorms
            sumw2 = sumw2 * binnorms**2

        if patch_opts is not None:
            opts = {'cmap': 'viridis'}
            opts.update(patch_opts)
            pc = ax.pcolormesh(xedges, yedges, sumw.T, **opts)
            ax.add_collection(pc)
            if clear:
                fig.colorbar(pc, ax=ax, label=hist.label)
        if text_opts is not None:
            for ix, xcenter in enumerate(xaxis.centers()):
                for iy, ycenter in enumerate(yaxis.centers()):
                    opts = {
                        'horizontalalignment': 'center',
                        'verticalalignment': 'center',
                    }
                    if patch_opts is not None:
                        opts['color'] = 'black' if pc.norm(sumw[ix, iy]) > 0.5 else 'lightgrey'
                    opts.update(text_opts)
                    txtformat = opts.pop('format', r'%.2g')
                    ax.text(xcenter, ycenter, txtformat % sumw[ix, iy], **opts)

    if clear:
        ax.set_xlabel(xaxis.label)
        ax.set_ylabel(yaxis.label)
        ax.set_xlim(xedges[0], xedges[-1])
        ax.set_ylim(yedges[0], yedges[-1])

    return ax


def plotgrid(h, figure=None, row=None, col=None, overlay=None, row_overflow='none', col_overflow='none', **plot_opts):
    """Create a grid of plots, enumerating identifiers on up to 3 axes

    Parameters
    ----------
        h : Hist
            A histogram with up to 3 axes
        figure : matplotlib.figure.Figure, optional
            If omitted, a new figure is created. Otherwise, the axes will be redrawn on this existing figure.
        row : str
            Name of row axis
        col : str
            Name of column axis
        overlay : str
            name of overlay axis
        row_overflow : str, optional
            If overflow behavior is not 'none', extra bins will be drawn on either end of the nominal x
            axis range, to represent the contents of the overflow bins.  See `Hist.sum` documentation
            for a description of the options.
        col_overflow : str, optional
            Similar to ``row_overflow``
        ``**plot_opts`` : kwargs
            The remaining axis of the histogram, after removing any of ``row,col,overlay`` specified,
            will be the plot axis, with ``plot_opts`` passed to the `plot1d` call.

    Returns
    -------
        axes : numpy.ndarray
            An array of matplotlib `Axes <https://matplotlib.org/3.1.1/api/axes_api.html>`_ objects
    """
    import mplhep as hep
    import matplotlib.pyplot as plt
    haxes = set(ax.name for ax in h.axes())
    nrow, ncol = 1, 1
    if row:
        row_identifiers = h.identifiers(row, overflow=row_overflow)
        nrow = len(row_identifiers)
        haxes.remove(row)
    if col:
        col_identifiers = h.identifiers(col, overflow=col_overflow)
        ncol = len(col_identifiers)
        haxes.remove(col)
    if overlay:
        haxes.remove(overlay)
    if len(haxes) > 1:
        raise ValueError("More than one dimension left: %s" % (",".join(ax for ax in haxes),))
    elif len(haxes) == 0:
        raise ValueError("Not enough dimensions available in %r" % h)

    figsize = plt.rcParams['figure.figsize']
    figsize = figsize[0] * max(ncol, 1), figsize[1] * max(nrow, 1)
    if figure is None:
        fig, axes = plt.subplots(nrow, ncol, figsize=figsize, squeeze=False, sharex=True, sharey=True)
    else:
        fig = figure
        shape = (0, 0)
        lastax = fig.get_children()[-1]
        if isinstance(lastax, plt.Axes):
            shape = lastax.rowNum + 1, lastax.colNum + 1
        if shape[0] == nrow and shape[1] == ncol:
            axes = np.array(fig.axes).reshape(shape)
        else:
            fig.clear()
            # fig.set_size_inches(figsize)
            axes = fig.subplots(nrow, ncol, squeeze=False, sharex=True, sharey=True)

    for icol in range(ncol):
        hcol = h
        coltitle = None
        if col:
            vcol = col_identifiers[icol]
            hcol = h.integrate(col, vcol)
            coltitle = str(vcol)
            if isinstance(vcol, Interval) and vcol.label is None:
                coltitle = "%s ∈ %s" % (h.axis(col).label, coltitle)
        for irow in range(nrow):
            ax = axes[irow, icol]
            hplot = hcol
            rowtitle = None
            if row:
                vrow = row_identifiers[irow]
                hplot = hcol.integrate(row, vrow)
                rowtitle = str(vrow)
                if isinstance(vrow, Interval) and vrow.label is None:
                    rowtitle = "%s ∈ %s" % (h.axis(row).label, rowtitle)

            plot1d(hplot, ax=ax, overlay=overlay, **plot_opts)
            if row is not None and col is not None:
                ax.set_title("%s, %s" % (rowtitle, coltitle))
            elif row is not None:
                ax.set_title(rowtitle)
            elif col is not None:
                ax.set_title(coltitle)

    for ax in axes.flatten():
        ax.autoscale(axis='y')
        ax.set_ylim(0, None)

    return axes



import numpy
import awkward1


def apply_global_index(index, target):
    def flat_take(layout):
        idx = awkward1.Array(layout)
        return target._content()[idx.mask[idx >= 0]]

    def descend(layout, depth):
        if layout.purelist_depth == 1:
            return lambda: flat_take(layout)

    (index,) = awkward1.broadcast_arrays(index)
    out = awkward1._util.recursively_apply(index.layout, descend)
    return awkward1._util.wrap(out, target.behavior)


def apply_local_index(index, target):
    index = index + target._starts()
    return apply_global_index(index, target)


def get_crossref(index, target):
    index = index.mask[index >= 0]
    return apply_local_index(index, target)



from __future__ import print_function, division

from coffea import hist
from coffea.util import numpy as np

from dummy_distributions import dummy_jagged_eta_pt
import pytest
import sys

def test_hist():
    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    h_nothing = hist.Hist("empty inside")
    assert h_nothing.sparse_dim() == h_nothing.dense_dim() == 0
    assert h_nothing.values() == {}

    h_regular_bins = hist.Hist("regular joe", hist.Bin("x", "x", 20, 0, 200), hist.Bin("y", "why", 20, -3, 3))
    h_regular_bins.fill(x=test_pt, y=test_eta)
    nentries = np.sum(counts)
    assert h_regular_bins.sum("x", "y", overflow='all').values(sumw2=True)[()] == (nentries, nentries)
    # bin x=2, y=10 (when overflow removed)
    count_some_bin = np.sum((test_pt>=20.)&(test_pt<30.)&(test_eta>=0.)&(test_eta<0.3))
    assert h_regular_bins.integrate("x", slice(20, 30)).values()[()][10] == count_some_bin
    assert h_regular_bins.integrate("y", slice(0, 0.3)).values()[()][2] == count_some_bin

    h_reduced = h_regular_bins[10:,-.6:]
    # bin x=1, y=2
    assert h_reduced.integrate("x", slice(20, 30)).values()[()][2] == count_some_bin
    assert h_reduced.integrate("y", slice(0, 0.3)).values()[()][1] == count_some_bin
    h_reduced.fill(x=23, y=0.1)
    assert h_reduced.integrate("x", slice(20, 30)).values()[()][2] == count_some_bin + 1
    assert h_reduced.integrate("y", slice(0, 0.3)).values()[()][1] == count_some_bin + 1

    animal = hist.Cat("animal", "type of animal")
    vocalization = hist.Cat("vocalization", "onomatopoiea is that how you spell it?")
    h_cat_bins = hist.Hist("I like cats", animal, vocalization)
    h_cat_bins.fill(animal="cat", vocalization="meow", weight=2.)
    h_cat_bins.fill(animal="dog", vocalization="meow", weight=np.array([-1., -1., -5.]))
    h_cat_bins.fill(animal="dog", vocalization="woof", weight=100.)
    h_cat_bins.fill(animal="dog", vocalization="ruff")
    assert h_cat_bins.values()[("cat", "meow")] == 2.
    assert h_cat_bins.values(sumw2=True)[("dog", "meow")] == (-7., 27.)
    assert h_cat_bins.integrate("vocalization", ["woof", "ruff"]).values(sumw2=True)[("dog",)] == (101., 10001.)

    height = hist.Bin("height", "height [m]", 10, 0, 5)
    h_mascots_1 = hist.Hist("fermi mascot showdown",
                          animal,
                          vocalization,
                          height,
                          # weight is a reserved keyword
                          hist.Bin("mass", "weight (g=9.81m/s**2) [kg]", np.power(10., np.arange(5)-1)),
                        )
    
    h_mascots_2 = hist.Hist("fermi mascot showdown",
                          axes=(animal,
                                vocalization,
                                height,
                                # weight is a reserved keyword
                                hist.Bin("mass", "weight (g=9.81m/s**2) [kg]", np.power(10., np.arange(5)-1)),)
                           )
                           
    h_mascots_3 = hist.Hist(
                         axes=[animal,
                               vocalization,
                               height,
                               # weight is a reserved keyword
                               hist.Bin("mass", "weight (g=9.81m/s**2) [kg]", np.power(10., np.arange(5)-1)),],
                            label="fermi mascot showdown"
                          )
                          
    h_mascots_4 = hist.Hist(
                            "fermi mascot showdown",
                            animal,
                            vocalization,
                            height,
                            # weight is a reserved keyword
                            hist.Bin("mass", "weight (g=9.81m/s**2) [kg]", np.power(10., np.arange(5)-1)),
                         axes=[animal,
                               vocalization,
                               height,
                               # weight is a reserved keyword
                               hist.Bin("mass", "weight (g=9.81m/s**2) [kg]", np.power(10., np.arange(5)-1)),],
                         
                       )
                          
    assert h_mascots_1._dense_shape == h_mascots_2._dense_shape
    assert h_mascots_2._dense_shape == h_mascots_3._dense_shape
    assert h_mascots_3._dense_shape == h_mascots_4._dense_shape
    
    assert h_mascots_1._axes == h_mascots_2._axes
    assert h_mascots_2._axes == h_mascots_3._axes
    assert h_mascots_3._axes == h_mascots_4._axes
                        
    adult_bison_h = np.random.normal(loc=2.5, scale=0.2, size=40)
    adult_bison_w = np.random.normal(loc=700, scale=100, size=40)
    h_mascots_1.fill(animal="bison", vocalization="huff", height=adult_bison_h, mass=adult_bison_w)
    goose_h = np.random.normal(loc=0.4, scale=0.05, size=1000)
    goose_w = np.random.normal(loc=7, scale=1, size=1000)
    h_mascots_1.fill(animal="goose", vocalization="honk", height=goose_h, mass=goose_w)
    crane_h = np.random.normal(loc=1, scale=0.05, size=4)
    crane_w = np.random.normal(loc=10, scale=1, size=4)
    h_mascots_1.fill(animal="crane", vocalization="none", height=crane_h, mass=crane_w)

    with pytest.raises(ValueError):
        h_mascots_1.fill(beast="crane", yelling="none", tallness=crane_h, heavitivity=crane_w)
    
    
    h_mascots_2 = h_mascots_1.copy()
    h_mascots_2.clear()
    baby_bison_h = np.random.normal(loc=.5, scale=0.1, size=20)
    baby_bison_w = np.random.normal(loc=200, scale=10, size=20)
    baby_bison_cutefactor = 2.5*np.ones_like(baby_bison_w)
    h_mascots_2.fill(animal="bison", vocalization="baa", height=baby_bison_h, mass=baby_bison_w, weight=baby_bison_cutefactor)
    h_mascots_2.fill(animal="fox", vocalization="none", height=1., mass=30.)

    h_mascots = h_mascots_1 + h_mascots_2
    assert h_mascots.integrate("vocalization", "h*").sum("height", "mass", "animal").values()[()] == 1040.

    species_class = hist.Cat("species_class", "where the subphylum is vertibrates")
    classes = {
        'birds': ['goose', 'crane'],
        'mammals': ['bison', 'fox'],
    }
    h_species = h_mascots.group("animal", species_class, classes)

    assert set(h_species.integrate("vocalization").values().keys()) == set([('birds',), ('mammals',)])
    nbirds_bin = np.sum((goose_h>=0.5)&(goose_h<1)&(goose_w>10)&(goose_w<100))
    nbirds_bin += np.sum((crane_h>=0.5)&(crane_h<1)&(crane_w>10)&(crane_w<100))
    assert h_species.integrate("vocalization").values()[('birds',)][1,2] == nbirds_bin
    tally = h_species.sum("mass", "height", "vocalization").values()
    assert tally[('birds',)] == 1004.
    assert tally[('mammals',)] == 91.

    h_species.scale({"honk": 0.1, "huff": 0.9}, axis="vocalization")
    h_species.scale(5.)
    tally = h_species.sum("mass", height, vocalization).values(sumw2=True)
    assert tally[('birds',)] == (520., 350.)
    assert tally[('mammals',)] == (435., 25*(40*(0.9**2)+20*(2.5**2)+1))

    assert h_species.axis("vocalization") is vocalization
    assert h_species.axis("height") is height
    assert h_species.integrate("vocalization", "h*").axis("height") is height

    tall_class = hist.Cat("tall_class", "species class (species above 1m)")
    mapping = {
        'birds': (['goose', 'crane'], slice(1., None)),
        'mammals': (['bison', 'fox'], slice(1., None)),
    }
    h_tall = h_mascots.group((animal, height), tall_class, mapping)
    tall_bird_count = np.sum(goose_h>=1.) + np.sum(crane_h>=1)
    assert h_tall.sum("mass", "vocalization").values()[('birds',)] == tall_bird_count
    tall_mammal_count = np.sum(adult_bison_h>=1.) + np.sum(baby_bison_h>=1) + 1
    assert h_tall.sum("mass", "vocalization").values()[('mammals',)] == tall_mammal_count

    h_less = h_mascots.remove(["fox", "bison"], axis="animal")
    assert h_less.sum("vocalization", "height", "mass", "animal").values()[()] == 1004.

def test_export1d():
    import uproot
    import os
    from coffea.hist import export1d
    

    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    h_regular_bins = hist.Hist("regular_joe", hist.Bin("x", "x", 20, 0, 200))
    h_regular_bins.fill(x=test_pt)

    hout = export1d(h_regular_bins)

    filename = 'test_export1d.root'
    
    with uproot.create(filename) as fout:
        fout['regular_joe'] = hout
        fout.close()

    with uproot.open(filename) as fin:
        hin = fin['regular_joe']


    assert(np.all(hin.edges == hout.edges))
    assert(np.all(hin.values == hout.values))

    del hin
    del fin

    if os.path.exists(filename):
        os.remove(filename)

def test_hist_serdes():
    import pickle
    h_regular_bins = hist.Hist("regular joe",
                               hist.Bin("x", "x", 20, 0, 200),
                               hist.Bin("y", "why", 20, -3, 3))

    h_regular_bins.fill(x=np.array([1.,2.,3.,4.,5.]),y=np.array([-2.,1.,0.,1.,2.]))

    h_regular_bins.sum('x').identifiers('y')

    spkl = pickle.dumps(h_regular_bins)
    
    hnew = pickle.loads(spkl)
    
    hnew.sum('x').identifiers('y')
    
    assert(h_regular_bins._dense_shape == hnew._dense_shape)
    assert(h_regular_bins._axes == hnew._axes)

def test_hist_serdes_labels():
    import pickle
    ax = hist.Bin('asdf', 'asdf', 3, 0, 3)
    ax.identifiers()[0].label = 'type 1'
    h = hist.Hist('a', ax)
    h.identifiers('asdf')

    spkl = pickle.dumps(h)
    
    hnew = pickle.loads(spkl)
    
    for old, new in zip(h.identifiers('asdf'), hnew.identifiers('asdf')):
        assert(old.label == new.label)

    assert(h._dense_shape == hnew._dense_shape)
    assert(h._axes == hnew._axes)

@pytest.mark.skipif(sys.version_info < (3, 4), reason="requires python3.4 or higher, test file is pickle proto 4")
def test_hist_compat():
    from coffea.util import load
    
    test = load('tests/samples/old_hist_format.coffea')
    
    expected_bins = np.array([ -np.inf, 0.,   20.,   40.,   60.,   80.,  100.,  120.,  140.,
                               160.,  180.,  200.,  220.,  240.,  260.,  280.,  300.,  320.,
                               340.,  360.,  380.,  400.,  420.,  440.,  460.,  480.,  500.,
                               520.,  540.,  560.,  580.,  600.,  620.,  640.,  660.,  680.,
                               700.,  720.,  740.,  760.,  780.,  800.,  820.,  840.,  860.,
                               880.,  900.,  920.,  940.,  960.,  980., 1000., 1020., 1040.,
                              1060., 1080., 1100., 1120., 1140., 1160., 1180., 1200.,   np.inf,
                              np.nan])
    assert np.all(test._axes[2]._interval_bins[:-1] == expected_bins[:-1])
    assert np.isnan(test._axes[2]._interval_bins[-1])

def test_issue_247():
    from coffea import hist

    h = hist.Hist('stuff', hist.Bin('old', 'old', 20, -1, 1))
    h.fill(old=h.axis('old').centers())
    h2 = h.rebin(h.axis('old'), hist.Bin('new', 'new', 10, -1, 1))
    # check first if its even possible to have correct binning
    assert np.all(h2.axis('new').edges() == h.axis('old').edges()[::2])
    # make sure the lookup works properly
    assert np.all(h2.values()[()] == 2.)
    h3 = h.rebin(h.axis('old'), 2)
    assert np.all(h3.values()[()] == 2.)

    with pytest.raises(ValueError):
        # invalid division
        _ = h.rebin(h.axis('old'), hist.Bin('new', 'new', 8, -1, 1))

    newaxis = hist.Bin('new', 'new', h.axis('old').edges()[np.cumsum([0, 2, 3, 5])])
    h4 = h.rebin('old', newaxis)
